#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/file.h>
#include <string.h>
#include <malloc.h>
#include "util.h"
#include "alloc.h"


#define TRUE 1
#define FALSE 0
//#include <iostream>
//using namespace std;

int main(int argc,char *argv[]) {  /* BEGIN MAIN */
	int argcc,mode=2,nrows,ncols;
	int cflag=FALSE,oflag=FALSE,fflag=FALSE;
	int i,NoOfFrames=8;
	double mse = 0;
	char OrgMovie[80],CompMovie[80];
	FILE *fOrgMovie,*fCompMovie;
	float *OrgFrame, *CompFrame, *DummyCrCbFrame, PSNR;
	double Average;
	
	for(argcc=1; argcc<argc; argcc++){
		if(argv[argcc][0] != '-'){
			fprintf(stderr,"Bad argument!");
			exit(1);
		}
		else{
			switch (argv[argcc][1]){
	      		case 'c':  /* compressed movie */
	        		strcpy(CompMovie,argv[++argcc]);
	        		cflag=TRUE;
					break;
				case 'o':  /* original movie */
					strcpy(OrgMovie,argv[++argcc]);
					oflag=TRUE;
					break;
				case 'm': /* type of sequence */
					mode = atoi(argv[++argcc]);
					break;
				case 'f': /*number of frames*/
					NoOfFrames = atoi(argv[++argcc]);
					fflag=TRUE;
					break;	      
				default:
					fprintf(stderr, "Wrong input parameter %c \n",argv[argcc][1]);
					break;
			}
		}
	}
	if (!(fflag && oflag && cflag)) {
		fprintf(stderr,"Not enough inputs \n");
		fprintf(stderr,"Input should contain: \n");
		fprintf(stderr,"-c COMPRESSEDMOVIE.YUV\n");
		fprintf(stderr,"-o ORIGINALMOVIE.YUV \n");
		fprintf(stderr,"-f NUMBEROFFRAMES \n");
		exit(1);
	}
	
    if ((fOrgMovie = fopen(OrgMovie,"rb")) == NULL){
    	fprintf(stderr,"Cannot open file %s for reading\n",OrgMovie);
		exit(1);
	}  	
	if ((fCompMovie = fopen(CompMovie,"rb")) == NULL) {
		printf("Cannot open file %s for reading\n",CompMovie);
		exit(1);	    
	}
	
	switch (mode){
		case 1:  /* qcif */
			nrows = 144;
			ncols = 176;
			break;
		case 2:  /* cif */
			nrows = 288;
			ncols = 352;
			break;
		case 3:  /* sif */
			nrows = 240;
			ncols = 352;
			break;
		case 4:
			nrows = 512;
			ncols = 512;
			break;
		default:
			perror("Illegal mode\n");
			exit(1);
	}
	
	OrgFrame = matrix1_f(nrows*ncols);
	CompFrame = matrix1_f(nrows*ncols);
	DummyCrCbFrame = matrix1_f(nrows/2*ncols/2);
	
	for (i=0; i<NoOfFrames; i++) {
		Getframe(nrows*ncols, OrgFrame, fOrgMovie);
		Getframe(nrows/2*ncols/2, DummyCrCbFrame, fOrgMovie);
		Getframe(nrows/2*ncols/2, DummyCrCbFrame, fOrgMovie);
		
		Getframe(nrows*ncols, CompFrame, fCompMovie);
		Getframe(nrows/2*ncols/2, DummyCrCbFrame, fCompMovie);
		Getframe(nrows/2*ncols/2, DummyCrCbFrame, fCompMovie);
		
		mse += PSNRcalc(nrows*ncols, OrgFrame, CompFrame);		
		//fprintf(stderr,"PSNR frame %i= %.2f \n",i,PSNR);
	}
	Average = mse/(nrows*ncols*NoOfFrames);
	PSNR = (float)(10.0*log10(255.0*255.0/Average));
	
	fprintf(stderr,"Average PSNR=%.2f \n\n",PSNR);
	printf("%.2f \n",PSNR);
	free(OrgFrame);
	free(CompFrame);
	free(DummyCrCbFrame);
	
	return 0;
}
