#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/file.h>
#include <malloc.h>

#include "util.h"

/***************************************************************************
* Function for computing PSNR for two frames                 
*--------------------------------------------------------------------------
* Input:     N    - frame size 
*            x    - first frame                                   
*            y    - second frame
* Return:    PSNR - PSNR                                      
*--------------------------------------------------------------------------
* First version:          Arild Fuldseth        1996-01-31          
***************************************************************************/
 
double PSNRcalc(int N,float *x,float *y){
 
  int i;  
  double diff,mse;
 
  mse = 0.0;
  for (i=0;i<N;i++){
    diff = (double)x[i] - (double)y[i];
    mse += diff*diff;
  }   
  return mse;
}

/***************************************************************************
* Function for reading one frame from file                 
*--------------------------------------------------------------------------
* Input:     N    - frame size 
*            fp   - file pointer
* Output:    x    - frame                                   
*--------------------------------------------------------------------------
* First version:          Arild Fuldseth        1996-01-31          
***************************************************************************/
 
void Getframe(int N,float *x, FILE *fp){
 
  int i;
  
  for (i=0;i<N;i++) x[i] = (float)fgetc(fp);
}

/**************************************************************************
* Function for computing the squared distance between two vectors x and y  
*              d = SUM [(x(i)-y(i))*(x(i)-y(i)]                            
*--------------------------------------------------------------------------
* Input:  x    - first vector                                              
*         y    - second vector                                             
*         dim  - vector dimension                                          
* Return: dist - squared distance                                          
*--------------------------------------------------------------------------
* Arild Fuldseth                                      1993-04-29            
***************************************************************************/
 			       
float Distance(float *x,float *y,int dim){
   
  int m;  
  float dist;

  /* compute squared error distance */ 
  
  dist = 0.0;
  for (m=0; m<dim; m++){
    dist += (x[m]-y[m])*(x[m]-y[m]);
  }
  return dist;
}

/***************************************************************************
* Function for writing one frame to file                 
*--------------------------------------------------------------------------
* Input:     N    - frame size 
*            x    - frame                                   
*            fp   - file pointer
*--------------------------------------------------------------------------
* First version:          Arild Fuldseth        1996-01-31          
***************************************************************************/
 
void Putframe(int N,float *x, FILE *fp){
 
  int i;
 
  for (i=0;i<N;i++) fputc((char)((int)(x[i]+128.0)),fp);
}
 
/***************************************************************************
* Function for reading bitstream to file                 
*--------------------------------------------------------------------------
* Input:     nbits  - number of bits 
*            fp     - file pointer
* Output:    bits   - bit stream                                   
*--------------------------------------------------------------------------
* First version:          Arild Fuldseth        1996-01-31          
***************************************************************************/

void Getsymb(int nsymb,float *symb,FILE *fp){

  int i;

  for (i=0;i<nsymb;i++) fscanf(fp,"%f\n",&symb[i]);

}
