#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include "alloc.h"
#include "EZW.h"
#include "list.h"
//#include "fifo.h"

int zeroes, ones;
unsigned char mask,input_byte;


void ezw_decode(float **Image,int RowsImage,int ColsImage,int levels,int NoOfSubbands,int frame,float LPQuant,float LPThreshold,unsigned char *Code,int *CodeCnt)
{
	int BPP;
	EZWelement *EZW;
	float mean;

	//init
	mean = 0.0;
	zeroes = 0; ones = 0;
	input_byte = 0;
	mask = 0;

	EZW = (EZWelement *)calloc(RowsImage*ColsImage/(NoOfSubbands*NoOfSubbands),sizeof(EZWelement));

	if(frame == 0){
		memcpy(&mean,&Code[*CodeCnt],sizeof(float));
		*CodeCnt += sizeof(float);
	}

	BPP = (int)Code[(*CodeCnt)++];
	if (BPP ==0) {//check to see if maxvalue<LPThreshold
		if ((int)Code[(*CodeCnt)++]	== 0) {
			return;
		}
	}
			

	EZW_decode(EZW,Image,BPP,RowsImage,ColsImage,NoOfSubbands,levels,LPQuant,LPThreshold,Code,CodeCnt);

	writeToImage(Image,EZW,RowsImage,ColsImage,NoOfSubbands,levels,mean);

	free(EZW);
}

void EZW_decode(EZWelement *EZW,float **Image,int BPP,int RowsImage,int ColsImage,int NoOfSubbands,int levels,float LPQuant,float LPThreshold,unsigned char *Code,int *CodeCnt)
{
	int antall_codede = 0;
	int tmp_bpp = BPP-1;
	int i;
	float Threshold=LPThreshold;
	
    //Init_EZW
	init_EZW(Image,EZW,RowsImage,ColsImage,NoOfSubbands,levels);
	
	if (BPP==0) {//do one dominant pass and one subordinate pass
	  	tmp_bpp=0;
	  	dominant_pass(EZW,RowsImage,ColsImage,NoOfSubbands,levels,Threshold,LPThreshold,Code,CodeCnt);
		subordinate_pass(EZW,Threshold,tmp_bpp,LPQuant,LPThreshold,Code,CodeCnt);		
	  	return;
  	}
	
	for (i=1;i<tmp_bpp;i++) {
		Threshold += (float)(1<<i)*LPQuant;
	}
	
	if (BPP==1)
		tmp_bpp=1;
	
	while (tmp_bpp>0) {	  	
	  	dominant_pass(EZW,RowsImage,ColsImage,NoOfSubbands,levels,Threshold,LPThreshold,Code,CodeCnt);

	  	antall_codede += list_length;
	  	subordinate_pass(EZW,Threshold,tmp_bpp,LPQuant,LPThreshold,Code,CodeCnt);
  	 
	  	tmp_bpp--;
	  	Threshold -= (float)(1<<tmp_bpp)*LPQuant;
	}
}

void dominant_pass(EZWelement *EZW,int RowsImage,int ColsImage,int NoOfSubbands,int levels,float threshold,float LPThreshold,unsigned char *Code,int *CodeCnt)
{
	int s;
	int tmp_number;
	int number=0;
	int i,j;
	int col_LP,row_LP;

	col_LP = (ColsImage/NoOfSubbands)>>levels;
	row_LP = (RowsImage/NoOfSubbands)>>levels;

	for (i=0; i<row_LP; i++) {
		for (j=0; j<col_LP; j++) {
			if (EZW[number].status == NC) {
				input_element(&EZW[number],RowsImage,ColsImage,NoOfSubbands,levels,Code,CodeCnt);
			}
			put_in_fifo(number);
			number++;
		}
	}
	s = get_from_fifo();
	while (fifo_empty==0) {
    	if (EZW[s].code!=ZTR && EZW[s].code!=AZTR) {
    		if (EZW[s].up_left != NULL) {
    			if (EZW[s].up_left->status == NC) {
    				input_element(EZW[s].up_left,RowsImage,ColsImage,NoOfSubbands,levels,Code,CodeCnt);
    			}
    			tmp_number = coeff2scan(EZW[s].up_left->pos_x,EZW[s].up_left->pos_y,RowsImage,ColsImage,NoOfSubbands,levels);
    			put_in_fifo(tmp_number);
    		}
    		if (EZW[s].up_right != NULL) {
	    		if (EZW[s].up_right->status == NC) {
	    			input_element(EZW[s].up_right,RowsImage,ColsImage,NoOfSubbands,levels,Code,CodeCnt);
	    		}
				tmp_number = coeff2scan(EZW[s].up_right->pos_x,EZW[s].up_right->pos_y,RowsImage,ColsImage,NoOfSubbands,levels);
    			put_in_fifo(tmp_number);
    		}
    		if (EZW[s].down_left != NULL) {
	    		if (EZW[s].down_left->status == NC) {
	    			input_element(EZW[s].down_left,RowsImage,ColsImage,NoOfSubbands,levels,Code,CodeCnt);
	    		}
	    		tmp_number = coeff2scan(EZW[s].down_left->pos_x,EZW[s].down_left->pos_y,RowsImage,ColsImage,NoOfSubbands,levels);
    			put_in_fifo(tmp_number);
    		}
    		if (EZW[s].down_right != NULL) {
	    		if (EZW[s].down_right->status == NC) {
		    		input_element(EZW[s].down_right,RowsImage,ColsImage,NoOfSubbands,levels,Code,CodeCnt);
	    		}
	    		tmp_number = coeff2scan(EZW[s].down_right->pos_x,EZW[s].down_right->pos_y,RowsImage,ColsImage,NoOfSubbands,levels);
    			put_in_fifo(tmp_number);
    		}
    	}
    	s = get_from_fifo();
	}
	destroy_fifo();
}

void subordinate_pass(EZWelement *EZW,float threshold,int BPP,float LPQuant,float LPThreshold,unsigned char *Code,int *CodeCnt) {
	list_type d;
	int quantvalue;
	float value;
	int i,j;
	char found;

	if (threshold>0) {
		for (i=0;i<list_length;i++) {
			quantvalue = 0;
			d = get_list_element(i,&found);
      		if (found==1) {
				for (j=0;j<BPP;j++) {
					quantvalue += ((int)get_bit(Code,CodeCnt))<<j;
				}
      		}
      		value = threshold+((float)quantvalue+0.5)*LPQuant;
      		if (EZW[d.x].code == POS)
      			EZW[d.x].value = value;
      		else if (EZW[d.x].code == NEG)
      			EZW[d.x].value = -value;
      		else {
      			fprintf(stderr,"error subordinate_pass,number=%i and code=%i \n",d.x,EZW[d.x].code);
      			exit(-1);
      		}
		}
	}
	destroy_list();
}

void input_element(EZWelement *EZW,int RowsImage,int ColsImage,int NoOfSubbands,int levels,unsigned char *Code,int *CodeCnt)
{
  list_type d;
  int number;
  int number_of_bits = 2;

  number = coeff2scan(EZW->pos_x,EZW->pos_y,RowsImage,ColsImage,NoOfSubbands,levels);

  d.x = number;
  EZW->code = input_code(number_of_bits,Code,CodeCnt);


  if ((EZW->code==POS)) {
//    m->m[s->y][s->x] = t;
    append_to_list(d);
    EZW->status = 1;
  }
  else if (EZW->code==NEG) {
    //m->m[s->y][s->x] = -t;
    append_to_list(d);
    EZW->status = 1;
  }
  else if (EZW->code==AZTR) {
  	set_state(EZW,1);  //this node and all children shall not be more decoded
  }
}

int input_code(int count,unsigned char *Code,int *CodeCnt) {
	int tmp=0;
	if (count == 1) {
		tmp = (int)get_bit(Code,CodeCnt);
		if (tmp) 	//1
			return(ONE);
		else 		//0
			return(ZERO);
	}
	tmp = (int)get_bit(Code,CodeCnt);
	if (tmp) {	//binary 1
		tmp = (int)get_bit(Code,CodeCnt);
		if (tmp)
			return (NEG);		//binary 11
		else
			return (POS);		//binary 10
	}
	else {				//binary 0
		tmp = (int)get_bit(Code,CodeCnt);
		if (tmp) {		// 01
			tmp = (int)get_bit(Code,CodeCnt);
			if (tmp)
				return (AZTR);	// 011
			else
				return (IZ);		// 010
		}
		else
			return (ZTR);			// 00
	}

	//should never come here
	return(0);



}

char get_bit(unsigned char *Code,int *CodeCnt)
{
  char bit;

  if (mask==0) {
    memcpy(&input_byte,&Code[*CodeCnt],sizeof(char));
    (*CodeCnt)++;
    mask = 0x80;
  }

  if ((input_byte&mask)==0) {
    bit = 0;
    zeroes++;
  }
  else {
    bit = 1;
    ones++;
  }

  mask >>= 1;

  return (bit);
}




void init_EZW(float **Image,EZWelement *EZW,int RowsImage,int ColsImage, int NoOfSubbands, int levels)
{
	int x,y;
	int number;
	int parent_number;
	for (y=0; y<RowsImage/NoOfSubbands; y++) {
		for (x=0; x<ColsImage/NoOfSubbands; x++) {
			number = coeff2scan(x,y,RowsImage,ColsImage,NoOfSubbands,levels);
			EZW[number].pos_x = x;
			EZW[number].pos_y = y;

			//set parent
			parent_number = get_parent(number,x,y,RowsImage,ColsImage,NoOfSubbands,levels);
			if (parent_number != -1)
				EZW[number].parent = &EZW[parent_number];
			else
				EZW[number].parent = NULL;

			//set children
			get_children(EZW,number,x,y,RowsImage,ColsImage,NoOfSubbands,levels);



			EZW[number].value = 0;
			EZW[number].status = NC; //not coded
		}
	}
}

void get_children(EZWelement *EZW,int number,int x,int y,int RowsImage,int ColsImage,int NoOfSubbands,int levels) {
	int child_nr;
	int child_x,child_y;
	int coeffs_in_LP=(RowsImage*ColsImage/(NoOfSubbands*NoOfSubbands))>>(2*levels);
	int col_level1,row_level1,col_level2,row_level2;

	col_level1 = (ColsImage/NoOfSubbands)>>1;
	row_level1 = (RowsImage/NoOfSubbands)>>1;
	col_level2 = col_level1>>1;
	row_level2 = row_level1>>1;

	if (levels ==3)	 {

	}
	else if (levels == 2) {
		//in LL2
		if (number < coeffs_in_LP) {
			//only 3 children
			EZW[number].up_left = NULL;

			//up right
			child_x = x+col_level2;
			child_y = y;
			child_nr = coeff2scan(child_x,child_y,RowsImage,ColsImage,NoOfSubbands,levels);
			EZW[number].up_right = &EZW[child_nr];

			//down left
			child_x = x;
			child_y = y+row_level2;
			child_nr = coeff2scan(child_x,child_y,RowsImage,ColsImage,NoOfSubbands,levels);
			EZW[number].down_left = &EZW[child_nr];

			//down right
			child_x = x+col_level2;
			child_y = y+row_level2;
			child_nr = coeff2scan(child_x,child_y,RowsImage,ColsImage,NoOfSubbands,levels);
			EZW[number].down_right = &EZW[child_nr];
		}

		//4 children

		//in HL2
		else if (number < 2*coeffs_in_LP) {
			child_x = col_level1 + (x-col_level2)*2;
			child_y = y*2;
			child_nr = coeff2scan(child_x,child_y,RowsImage,ColsImage,NoOfSubbands,levels);
			EZW[number].up_left = &EZW[child_nr];
			EZW[number].up_right = &EZW[child_nr+1];
			EZW[number].down_left = &EZW[child_nr+2];
			EZW[number].down_right = &EZW[child_nr+3];
		}
		//in LH2
		else if (number < 3*coeffs_in_LP) {
			child_x = 2*x;
			child_y = row_level1 + (y-row_level2)*2;
			child_nr = coeff2scan(child_x,child_y,RowsImage,ColsImage,NoOfSubbands,levels);
			EZW[number].up_left = &EZW[child_nr];
			EZW[number].up_right = &EZW[child_nr+1];
			EZW[number].down_left = &EZW[child_nr+2];
			EZW[number].down_right = &EZW[child_nr+3];
		}
		else if (number < 4*coeffs_in_LP) {
			child_x = col_level1 + (x-col_level2)*2;
			child_y = row_level1 + (y-row_level2)*2;
			child_nr = coeff2scan(child_x,child_y,RowsImage,ColsImage,NoOfSubbands,levels);
			EZW[number].up_left = &EZW[child_nr];
			EZW[number].up_right = &EZW[child_nr+1];
			EZW[number].down_left = &EZW[child_nr+2];
			EZW[number].down_right = &EZW[child_nr+3];
		}
		//in HL1, LH1 or HH1  - no children
		else {
			EZW[number].up_left = NULL;
			EZW[number].up_right = NULL;
			EZW[number].down_left = NULL;
			EZW[number].down_right = NULL;
		}
	}
	else {

	}
}


int get_parent(int number,int x,int y,int RowsImage,int ColsImage,int NoOfSubbands,int levels)
{
	int parent_nr;
	int coeffs_in_LP=(RowsImage*ColsImage/(NoOfSubbands*NoOfSubbands))>>(2*levels);
	int col_level1,row_level1,col_level2,row_level2;

	col_level1 = (ColsImage/NoOfSubbands)/2;
	row_level1 = (RowsImage/NoOfSubbands)/2;
	col_level2 = col_level1/2;
	row_level2 = row_level1/2;


	//in LP-LP band, no parent
	if (number < coeffs_in_LP) {
		return -1; //no parent
	}

	//in same dyadic split level
	else if (number < 4*coeffs_in_LP) {
		if (x >= col_level2)
			x-= col_level2;
		if (y >= row_level2)
			y -= row_level2;
		parent_nr = coeff2scan(x,y,RowsImage,ColsImage,NoOfSubbands,levels);
	}
	else {
		coeffs_in_LP <<=2;
		x = x/2;
		y = y/2;
		parent_nr = coeff2scan(x,y,RowsImage,ColsImage,NoOfSubbands,levels);
	}
	return parent_nr;
}

void set_state(EZWelement *EZW,unsigned char state) {
	EZW->status = state;
	if (EZW->up_left != NULL)
		set_state(EZW->up_left,state);
	if (EZW->up_right != NULL)
		set_state(EZW->up_right,state);
	if (EZW->down_left != NULL)
		set_state(EZW->down_left,state);
	if (EZW->down_right != NULL)
		set_state(EZW->down_right,state);
}


// from x,y to scan_number
int coeff2scan(int x, int y, int RowsImage, int ColsImage, int NoOfSubbands, int levels) {
	int number=0;
	int coeff_pr_parent;
	int pix_band3, pix_band2,pix_band1;
	int coeff_in_block_col;
	int pix_LPLP = RowsImage*ColsImage/(NoOfSubbands*NoOfSubbands);
	int cols_pr_band_level_one,rows_pr_band_level_one;
	int cols_pr_band_level_two,rows_pr_band_level_two;

	pix_band1 = pix_LPLP/4;

	cols_pr_band_level_one = (ColsImage/NoOfSubbands)/2;
	rows_pr_band_level_one = (RowsImage/NoOfSubbands)/2;

	cols_pr_band_level_two = cols_pr_band_level_one/2;
	rows_pr_band_level_two = rows_pr_band_level_one/2;

	if (levels == 3) {
		fprintf(stderr,"coeff2scan not implementet for level=3");
		exit(1);

	}
	else if (levels == 2) {
		pix_band2 = pix_band1/4;
		//in dyadic split region
		if (x<cols_pr_band_level_one && y<rows_pr_band_level_one) {
			//in LL2
			if (x<cols_pr_band_level_two && y<rows_pr_band_level_two) {
				number = y*cols_pr_band_level_two + x;
			}
			//in HL2
			else if (x >= cols_pr_band_level_two && y<rows_pr_band_level_two) {
				number = pix_band2;
				x -= cols_pr_band_level_two;
				number += y*cols_pr_band_level_two + x;
			}
			//in LH2
			else if (x<cols_pr_band_level_two && y >= rows_pr_band_level_two) {
				number = 2*pix_band2;
				y -= rows_pr_band_level_two;
				number += y*cols_pr_band_level_two + x;
			}
			//in HH2
			else {
				number = 3*pix_band2;
				x -= cols_pr_band_level_two;
				y -= rows_pr_band_level_two;
				number += y*cols_pr_band_level_two + x;
			}
			return number;
		}
		//in HL1,LH1 or HH1
		else {
			coeff_pr_parent = 4;
			coeff_in_block_col = cols_pr_band_level_one/2*coeff_pr_parent;
			//in HL1
			if (x >= cols_pr_band_level_one && y<rows_pr_band_level_one) {
				number = pix_band1;
				x -= cols_pr_band_level_one;
				number += (y/2)*coeff_in_block_col + (x/2)*coeff_pr_parent + (y%2)*2 + (x%2);
			}
			//in LH1
			else if (x<cols_pr_band_level_one && y >= rows_pr_band_level_one) {
				number = 2*pix_band1;
				y -= rows_pr_band_level_one;
				number += (y/2)*coeff_in_block_col + (x/2)*coeff_pr_parent + (y%2)*2 + (x%2);
			}
			//in HH1
			else {
				number = 3*pix_band1;
				y -= rows_pr_band_level_one;
				x -= cols_pr_band_level_one;
				number += (y/2)*coeff_in_block_col + (x/2)*coeff_pr_parent + (y%2)*2 + (x%2);
			}
			return number;
		}
	}
	else { //level = 1
		pix_band3 = pix_band2/4;
		fprintf(stderr,"coeff2scan not implementet for level=1");
		exit(1);
	}
	return number;



}


void scan2coeff(int number,int *x,int *y, int RowsImage,int ColsImage,int NoOfSubbands,int levels) {
	int pix_band3, pix_band2,pix_band1;
	int rows_pr_band,cols_pr_band;
	int start_x,start_y;
	int coeff_pr_parent;
	int block_nr,blocks_pr_col;
	int pix_LPLP = RowsImage*ColsImage/(NoOfSubbands*NoOfSubbands);

	pix_band1 = pix_LPLP/4;
	if (levels>1)
		pix_band2 = pix_band1/4;
	else
		pix_band2 = 0;
	if (levels>2)
		pix_band3 = pix_band2/4;
	else
		pix_band3 = 0;

	if (levels == 3) {
		fprintf(stderr,"scan2coeff not implemented for level=3");
		exit(1);



	}

	else if (levels == 2) {
		rows_pr_band = RowsImage>>levels;
		cols_pr_band = ColsImage>>levels;

		//in LL2
		if (number<pix_band2) {
			*x = number%cols_pr_band;
			*y = number/cols_pr_band;
		}
		//in HL2
		else if (number<2*pix_band2) {
			*x = cols_pr_band + (number-pix_band2)%cols_pr_band;
			*y = (number-pix_band2)/cols_pr_band;
		}
		//in LH2
		else if (number<3*pix_band2) {
			*x = (number-2*pix_band2)%cols_pr_band;
			*y = rows_pr_band + (number-2*pix_band2)/cols_pr_band;
		}
		//in HH2
		else if (number<pix_band1) {
			*x = cols_pr_band + (number-3*pix_band2)%cols_pr_band;
			*y = rows_pr_band + (number-3*pix_band2)/cols_pr_band;
		}

		else {
			coeff_pr_parent = 4;
			rows_pr_band = RowsImage>>(levels-1);
			cols_pr_band = ColsImage>>(levels-1);

			//blocks_pr_col = bredde HL1 /2 (to coeff pr blokk i bredden)
			blocks_pr_col = (ColsImage/NoOfSubbands)/4;
			//in HL1
			if (number<2*pix_band1) {

				start_x = (ColsImage/NoOfSubbands)/2;
				start_y = 0;
				number -= pix_band1;

				//START HER!!!!!!
				block_nr = number/coeff_pr_parent;
				start_x += block_nr%blocks_pr_col;
				start_y += 2*block_nr/blocks_pr_col;

				number -= coeff_pr_parent*(number/coeff_pr_parent);
				if (number == 1)
					start_x++;
				else if (number == 2)
					start_y++;
				else if (number == 3) {
					start_x++;
					start_y++;
				}
				*x = start_x;
				*y = start_y;
			}
			//in LH1
			else if (number<3*pix_band1) {
				start_x = 0;
				start_y = (RowsImage/NoOfSubbands)/2;
				number -= 2*pix_band1;

				block_nr = number/coeff_pr_parent;
				start_x += block_nr%blocks_pr_col;
				start_y += 2*block_nr/blocks_pr_col;

				number -= coeff_pr_parent*(number/coeff_pr_parent);
				if (number == 1)
					start_x++;
				else if (number == 2)
					start_y++;
				else if (number == 3) {
					start_x++;
					start_y++;
				}
				*x = start_x;
				*y = start_y;
			}
			//in HH1
			else {
				start_x = (ColsImage/NoOfSubbands)/2;
				start_y = (RowsImage/NoOfSubbands)/2;
				number -= 3*pix_band1;

				block_nr = number/coeff_pr_parent;
				start_x += block_nr%blocks_pr_col;
				start_y += 2*block_nr/blocks_pr_col;

				number -= coeff_pr_parent*(number/coeff_pr_parent);
				if (number == 1)
					start_x++;
				else if (number == 2)
					start_y++;
				else if (number == 3) {
					start_x++;
					start_y++;
				}
				*x = start_x;
				*y = start_y;
			}
		}
	}

	else {
		fprintf(stderr,"scan2coeff not implemented for level=1");
		exit(1);
	}
}

void writeToImage(float **Image,EZWelement *EZW,int RowsImage,int ColsImage,int NoOfSubbands,int levels,float mean)
{
	int y,x;
	int number;
	int col_LP,row_LP;
	for (y=0;y<RowsImage/NoOfSubbands;y++) {
		for (x=0;x<ColsImage/NoOfSubbands;x++) {
			number = coeff2scan(x,y,RowsImage,ColsImage,NoOfSubbands,levels);
			Image[y][x] = EZW[number].value;
		}
	}
	if (mean != 0.0) {
		col_LP = (ColsImage/NoOfSubbands)>>levels;
		row_LP = (RowsImage/NoOfSubbands)>>levels;

		for (y=0;y<row_LP;y++) {
			for (x=0;x<col_LP;x++) {
				Image[y][x] += mean;
			}
		}
	}
}
