#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include "codecdecl.h"
#include "util.h"
#include "alloc.h"
#include "filtering_k.h"

//int ImageDecode(unsigned char *Code,unsigned char *SideInfo,struct DecoderArgs *Args,FILE *file,float *Output,int *CodeBuffer,int *DestBuffer,unsigned char **Arithcode,unsigned char *QuantIndexArray1D)
int ImageDecode(float **ImageArray,unsigned char **Code,int CodeCnt,struct DecoderArgs *Args,FILE *file,CABAC_DecoderEnvironment *CodingEnvironment)
{  /* BEGIN MAIN */

	short int **QuantIndexArrayY;	
	int i, j,k,l;
	int frame;
	int RowsImage, ColsImage, FileSize,levels;
	int NoOfSubbands; 	
	int CodeSize=MAXCBITS;
	int buffer;
	float QuantStepSize, Threshold;
	float LPQuant, LPThreshold;
	float **SubbandImageY;
	char TmpName[40];	
	
	MacroBlock *mb;
	
	//CABAC variables
	int model_number, img_type, ctx_number;
	DecodingEnvironment* dep;
	MotionInfoContexts* mot_ctx;
	TextureInfoContexts* tex_ctx;

	/* Set default values */	
	Threshold=(float)Args->threshold;
	QuantStepSize = (float)Args->quality;
	RowsImage = Args->height;
	ColsImage = Args->width;
	NoOfSubbands=NOOFSUBBANDSLUMIN;
	buffer = Args->buffer;	
	levels = (int)Args->levels;
	LPQuant = Args->LPquant;
	LPThreshold = Args->LPthreshold;
	CodeCnt = 0;
	
	QuantIndexArrayY=matrix2_si(RowsImage,ColsImage);
	SubbandImageY=matrix2_f(RowsImage,ColsImage);
	mb = (MacroBlock *)calloc(RowsImage*ColsImage/(NoOfSubbands*NoOfSubbands),sizeof(MacroBlock));
	*Code=matrix1_uch(CodeSize);	
	
	//CABAC - set environments
	dep = &(CodingEnvironment->partArr)->de_cabac;
	img_type = 2;
	ctx_number = 0;
	mot_ctx = CodingEnvironment->mot_ctx; //ContextsModels
	tex_ctx = CodingEnvironment->tex_ctx;
	
	
	for (frame=0; frame<buffer; frame++) {
		//read from file;	
		FileSize = fread(Code[0], sizeof(unsigned char), CodeSize, file);
		
		//Decode LP-LP band (coded with EZW)
		ezw_decode(SubbandImageY,RowsImage,ColsImage,levels,NoOfSubbands,frame,LPQuant,LPThreshold,*Code,&CodeCnt);
//		fprintf(stderr, "CodeCnt after ezw= %i \n",CodeCnt);
		
//		sprintf(TmpName,"ezw_dec%i.txt",frame);
//		writeTxtFile2_f(TmpName,SubbandImageY,RowsImage/8,ColsImage/8);
		
	
// 		******************************************************************
// 		* CABAC DECODING													*
// 		*																*
// 		******************************************************************		
		arideco_start_decoding(dep,*Code,CodeCnt,&CodeCnt,0);
		model_number = SetCtxModelNumber(0,img_type,ctx_number);
		
		//init contexts
		init_contexts(mot_ctx,tex_ctx,model_number,frame,buffer,QuantStepSize);
		
		//read macroblock
		read_macroblock(mb,CodingEnvironment,RowsImage,ColsImage,NoOfSubbands);
		
		//arideco done encoding
		arideco_done_decoding(dep);
		
		
		
		//inverse zig-zag scan
		inv_zig_zag_scan(mb,QuantIndexArrayY,RowsImage,ColsImage,NoOfSubbands);
		
//		sprintf(TmpName,"DecodeQuantIndexArray%i.txt",frame);
//		writeTxtFile2_si(TmpName, QuantIndexArrayY, RowsImage, ColsImage);
		
		
		/* Inverse uniform quantization */		
		invunifquant(SubbandImageY,QuantIndexArrayY,RowsImage,ColsImage,NoOfSubbands,QuantStepSize,Threshold);
//		sprintf(TmpName,"InvQuantIndexArray%i.txt",frame);
//		writeTxtFile2_f(TmpName, SubbandImageY, RowsImage, ColsImage);

		

		l=0;
		for(i=0; i<RowsImage; i++) {
			for (j=0; j<ColsImage; j++) {
				ImageArray[l++][frame] = SubbandImageY[i][j];
			}
		}
		//fprintf(stderr, "Decoded Codesize=%i \n",CodeCnt);	
		
		//set correct file-location		
		Args->codesize += CodeCnt;
		k=fseek(file, Args->codesize, SEEK_SET);			
		if (k!=0) {
			fprintf(stderr, "Position error in Codefile, position should be %lu but is %lu\n",Args->codesize,ftell(file));
			exit(ERROR);
		}
//		fprintf(stderr,"ftell(file) = %lu \n",ftell(file));
		CodeCnt = 0;
		
		//reset MB struct
		mb = (MacroBlock *)memset(mb,0,RowsImage*ColsImage/(NoOfSubbands*NoOfSubbands)*sizeof(MacroBlock));
	}
	freematrix2_f(SubbandImageY, RowsImage, ColsImage);
	freematrix2_si(QuantIndexArrayY,RowsImage,ColsImage);
	free(mb);
	free(*Code);
	return(0);

}  /* END MAIN */


void Read_All_Decoder_Parameters(struct DecoderArgs *Args,unsigned char *Code,int *CodeCnt){
	unsigned char tmpBuffer;
	
	memcpy(&Args->height,&Code[*CodeCnt],sizeof(int));
	*CodeCnt+=sizeof(int);
	memcpy(&Args->width,&Code[*CodeCnt],sizeof(int));
	*CodeCnt+=sizeof(int);
	memcpy(&Args->quality,&Code[*CodeCnt],sizeof(float));
	*CodeCnt+=sizeof(float);
	memcpy(&Args->threshold,&Code[*CodeCnt],sizeof(float));
	*CodeCnt+=sizeof(float);
	memcpy(&Args->height_old,&Code[*CodeCnt],sizeof(int));
	*CodeCnt+=sizeof(int);
	memcpy(&Args->width_old,&Code[*CodeCnt],sizeof(int));
	*CodeCnt+=sizeof(int);		
	memcpy(&Args->TotNoOfFrames,&Code[*CodeCnt],sizeof(int));
	*CodeCnt+=sizeof(int);	
	memcpy(&tmpBuffer,&Code[*CodeCnt],sizeof(char));
	*CodeCnt+=sizeof(char);
	Args->buffer = (int)tmpBuffer;
	memcpy(&Args->levels,&Code[*CodeCnt],sizeof(char));
	*CodeCnt+=sizeof(char);	
	memcpy(&Args->LPquant,&Code[*CodeCnt],sizeof(float));
	*CodeCnt+=sizeof(float);
	memcpy(&Args->LPthreshold,&Code[*CodeCnt],sizeof(float));
	*CodeCnt+=sizeof(float);
}
