
/*!
 *************************************************************************************
 * \file cabac.c
 *
 * \brief
 *    CABAC entropy coding routines
 *
 * \author
 *    Main contributors (see contributors.h for copyright, address and affiliation details)
 *    - Detlev Marpe                    <marpe@hhi.de>
 **************************************************************************************
 */

#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include "codecdecl.h"
#include "alloc.h"

int symbolCount = 0;
int last_dquant = 0;


/*!
 ************************************************************************
 * \brief
 *    This function is used to arithmetically decode the macroblock
 *    type info of a given MB.
 ************************************************************************
 */
//void readMB_skip_flagInfo_CABAC(SyntaxElement *se,struct inp_par *inp,struct img_par *img,DecodingEnvironment* dep_dp)
void readMB_skip_flagInfo_CABAC(DecodingEnvironment* dep_dp,MacroBlock* mb, MotionInfoContexts* ctx)
{
  int a, b;
  int act_ctx;
//  int bframe=(img->type==B_SLICE);    

  
    if (mb->mb_available_up == NULL)
      b = 0;
    else
      b = (( (mb->mb_available_up)->has_none_zero_coeffs == 0) ? 1 : 0 );
    if (mb->mb_available_left == NULL)
      a = 0;
    else
      a = (( (mb->mb_available_left)->has_none_zero_coeffs == 0) ? 1 : 0 );

    act_ctx = a + b;

    if (biari_decode_symbol(dep_dp, &ctx->mb_type_contexts[1][act_ctx]) == 1)
      mb->has_none_zero_coeffs = 0; //equals no significant coeffs
    else
      mb->has_none_zero_coeffs = 1;
  


//#if TRACE
//  fprintf(p_trace, "@%d %s\t\t%d\t%d %d\n",symbolCount++, se->tracestring, se->value1,a,b);
//  fflush(p_trace);
//#endif
//  if (!se->value1)
//  {
//    last_dquant=0;
//  }
  return;
}


static const int maxpos       [] = {16, 15, 64, 32, 32, 16,  4, 15,  8, 16};
static const int c1isdc       [] = { 1,  0,  1,  1,  1,  1,  1,  0,  1,  1};

static const int type2ctx_bcbp[] = { 0,  1,  2,  2,  3,  4,  5,  6,  5,  5}; // 7
static const int type2ctx_map [] = { 0,  1,  2,  3,  4,  5,  6,  7,  6,  6}; // 8
static const int type2ctx_last[] = { 0,  1,  2,  3,  4,  5,  6,  7,  6,  6}; // 8
static const int type2ctx_one [] = { 0,  1,  2,  3,  3,  4,  5,  6,  5,  5}; // 7
static const int type2ctx_abs [] = { 0,  1,  2,  3,  3,  4,  5,  6,  5,  5}; // 7
static const int max_c2       [] = { 4,  4,  4,  4,  4,  4,  3,  4,  3,  3}; // 9


//===== position -> ctx for MAP =====
//--- zig-zag scan ----
static const int  pos2ctx_map8x8 [] = { 0,  1,  2,  3,  4,  5,  5,  4,  4,  3,  3,  4,  4,  4,  5,  5,
                                        4,  4,  4,  4,  3,  3,  6,  7,  7,  7,  8,  9, 10,  9,  8,  7,
                                        7,  6, 11, 12, 13, 11,  6,  7,  8,  9, 14, 10,  9,  8,  6, 11,
                                       12, 13, 11,  6,  9, 14, 10,  9, 11, 12, 13, 11 ,14, 10, 12, 14}; // 15 CTX
static const int  pos2ctx_map8x4 [] = { 0,  1,  2,  3,  4,  5,  7,  8,  9, 10, 11,  9,  8,  6,  7,  8,
                                        9, 10, 11,  9,  8,  6, 12,  8,  9, 10, 11,  9, 13, 13, 14, 14}; // 15 CTX
static const int  pos2ctx_map4x4 [] = { 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 14}; // 15 CTX
static const int  pos2ctx_map2x4c[] = { 0,  0,  1,  1,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2}; // 15 CTX
static const int  pos2ctx_map4x4c[] = { 0,  0,  0,  0,  1,  1,  1,  1,  2,  2,  2,  2,  2,  2,  2,  2}; // 15 CTX
static const int* pos2ctx_map    [] = {pos2ctx_map4x4, pos2ctx_map4x4, pos2ctx_map8x8, pos2ctx_map8x4,
                                       pos2ctx_map8x4, pos2ctx_map4x4, pos2ctx_map4x4, pos2ctx_map4x4,
                                       pos2ctx_map2x4c, pos2ctx_map4x4c};
//--- interlace scan ----
//taken from ABT
static const int  pos2ctx_map8x8i[] = { 0,  1,  1,  2,  2,  3,  3,  4,  5,  6,  7,  7,  7,  8,  4,  5,
                                        6,  9, 10, 10,  8, 11, 12, 11,  9,  9, 10, 10,  8, 11, 12, 11,
                                        9,  9, 10, 10,  8, 11, 12, 11,  9,  9, 10, 10,  8, 13, 13,  9,
                                        9, 10, 10,  8, 13, 13,  9,  9, 10, 10, 14, 14, 14, 14, 14, 14}; // 15 CTX
static const int  pos2ctx_map8x4i[] = { 0,  1,  2,  3,  4,  5,  6,  3,  4,  5,  6,  3,  4,  7,  6,  8,
                                        9,  7,  6,  8,  9, 10, 11, 12, 12, 10, 11, 13, 13, 14, 14, 14}; // 15 CTX
static const int  pos2ctx_map4x8i[] = { 0,  1,  1,  1,  2,  3,  3,  4,  4,  4,  5,  6,  2,  7,  7,  8,
                                        8,  8,  5,  6,  9, 10, 10, 11, 11, 11, 12, 13, 13, 14, 14, 14}; // 15 CTX
//static const int* pos2ctx_map_int[] = {pos2ctx_map4x4, pos2ctx_map4x4, pos2ctx_map8x8i,pos2ctx_map8x4i,
//                                       pos2ctx_map4x8i,pos2ctx_map4x4, pos2ctx_map4x4, pos2ctx_map4x4,
//                                       pos2ctx_map2x4c, pos2ctx_map4x4c};


//===== position -> ctx for LAST =====
static const int  pos2ctx_last8x8 [] = { 0,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,
                                         2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,
                                         3,  3,  3,  3,  3,  3,  3,  3,  4,  4,  4,  4,  4,  4,  4,  4,
                                         5,  5,  5,  5,  6,  6,  6,  6,  7,  7,  7,  7,  8,  8,  8,  8}; //  9 CTX
static const int  pos2ctx_last8x4 [] = { 0,  1,  1,  1,  1,  1,  1,  1,  2,  2,  2,  2,  2,  2,  2,  2,
                                         3,  3,  3,  3,  4,  4,  4,  4,  5,  5,  6,  6,  7,  7,  8,  8}; //  9 CTX

static const int  pos2ctx_last4x4 [] = { 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15}; // 15 CTX
static const int  pos2ctx_last2x4c[] = { 0,  0,  1,  1,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2}; // 15 CTX
static const int  pos2ctx_last4x4c[] = { 0,  0,  0,  0,  1,  1,  1,  1,  2,  2,  2,  2,  2,  2,  2,  2}; // 15 CTX
static const int* pos2ctx_last    [] = {pos2ctx_last4x4, pos2ctx_last4x4, pos2ctx_last8x8, pos2ctx_last8x4,
                                        pos2ctx_last8x4, pos2ctx_last4x4, pos2ctx_last4x4, pos2ctx_last4x4,
                                        pos2ctx_last2x4c, pos2ctx_last4x4c};





/*!
 ************************************************************************
 * \brief
 *    Read Significance MAP
 ************************************************************************
 */
int read_significance_map (MacroBlock* mb,DecodingEnvironment* dep_dp,TextureInfoContexts* tex_ctx, int type, short int coeff[])
{
	int   i, sig;  
	int   i0        = 1; //LP-LP coded individually
	int   i1        = maxpos[type]-1;
	int   coeff_ctr = 0;

	BiContextType *map_ctx,*last_ctx;
  
	map_ctx = tex_ctx->map_contexts[type2ctx_map [type]];	
	last_ctx = tex_ctx->last_contexts[type2ctx_last[type]];
	
	for (i=i0; i<i1; i++) // if last coeff is reached, it has to be significant
  	{
    	//--- read significance symbol ---    
		sig = biari_decode_symbol   (dep_dp, map_ctx + pos2ctx_map     [type][i]);
		if (sig)
		{		 //i-1 because of LP-LP band
			coeff[i-1] = 1;
			coeff_ctr++;
			//--- read last coefficient symbol ---
			if (biari_decode_symbol (dep_dp, last_ctx + pos2ctx_last[type][i]))
			{
				//last sign.coeff = true, set rest of coeffs to 0
				for (i++; i<i1+1; i++) coeff[i-1] = 0;
			}
		}
		else
		{
			coeff[i-1] = 0;
		}
	}
	//--- last coefficient must be significant if no last symbol was received ---
	if (i<i1+1)
	{
		coeff[i-1] = 1;
		coeff_ctr++;
	}

	return coeff_ctr;
}



/*!
 ************************************************************************
 * \brief
 *    Read Levels
 ************************************************************************
 */
void read_significant_coefficients (MacroBlock* mb, DecodingEnvironment* dep_dp,TextureInfoContexts* tex_ctx,int type, short int coeff[])
{
	int   i, ctx;
	int   c1 = 1;
	int   c2 = 0;

	for (i=maxpos[type]-1; i>=1; i--)
	{
		if (coeff[i-1]!=0)
		{
			ctx = min (c1,4);
			coeff[i-1] += biari_decode_symbol (dep_dp, tex_ctx->one_contexts[type2ctx_one[type]] + ctx);
			if (coeff[i-1]==2)
			{
				ctx = min (c2, max_c2[type]);
				coeff[i-1] += unary_exp_golomb_level_decode (dep_dp, tex_ctx->abs_contexts[type2ctx_abs[type]]+ctx);
				c1=0;
				c2++;
			}
			else if (c1)
			{
				c1++;
			}
			if (biari_decode_symbol_eq_prob(dep_dp))
			{
				coeff[i-1] *= -1;
			}
		}
	}
}



/*!
 ************************************************************************
 * \brief
 *    decoding of unary binarization using one or 2 distinct
 *    models for the first and all remaining bins; no terminating
 *    "0" for max_symbol
 ***********************************************************************
 */
unsigned int unary_bin_max_decode(DecodingEnvironment* dep_dp,
                                  BiContextType* ctx,
                                  int ctx_offset,
                                  unsigned int max_symbol)
{
  unsigned int l;
  unsigned int symbol;
  BiContextType* ictx;

  symbol =  biari_decode_symbol(dep_dp, ctx );

  if (symbol==0)
    return 0;
  else
  {
    if (max_symbol == 1)
    return symbol;
    symbol=0;
    ictx=ctx+ctx_offset;
    do
    {
      l=biari_decode_symbol(dep_dp, ictx);
      symbol++;
    }
    while( (l!=0) && (symbol<max_symbol-1) );
    if ((l!=0) && (symbol==max_symbol-1))
      symbol++;
    return symbol;
  }
}


/*!
 ************************************************************************
 * \brief
 *    decoding of unary binarization using one or 2 distinct
 *    models for the first and all remaining bins
 ***********************************************************************
 */
unsigned int unary_bin_decode(DecodingEnvironment* dep_dp,
                              BiContextType* ctx,
                              int ctx_offset)
{
  unsigned int l;
  unsigned int symbol;
  BiContextType* ictx;

  symbol = biari_decode_symbol(dep_dp, ctx );

  if (symbol==0)
    return 0;
  else
  {
    symbol=0;
    ictx=ctx+ctx_offset;
    do
    {
      l=biari_decode_symbol(dep_dp, ictx);
      symbol++;
    }
    while( l!=0 );
    return symbol;
  }
}


/*!
 ************************************************************************
 * \brief
 *    finding end of a slice in case this is not the end of a frame
 *
 * Unsure whether the "correction" below actually solves an off-by-one
 * problem or whether it introduces one in some cases :-(  Anyway,
 * with this change the bit stream format works with CABAC again.
 * StW, 8.7.02
 ************************************************************************
 */
//int cabac_startcode_follows(struct img_par *img, struct inp_par *inp, int eos_bit)
//{
//  Slice         *currSlice  = img->currentSlice;
//  int           *partMap    = assignSE2partition[currSlice->dp_mode];
//  DataPartition *dP;
//  unsigned int  bit;
//  DecodingEnvironmentPtr dep_dp;
//  
//  dP = &(currSlice->partArr[partMap[SE_MBTYPE]]);
//  dep_dp = &(dP->de_cabac);
//
//  if( eos_bit )
//  {
//    bit = biari_decode_final (dep_dp); //GB
//    
//#if TRACE
//    //  strncpy(se->tracestring, "Decode Sliceterm", TRACESTRING_SIZE);
//    fprintf(p_trace, "@%d %s\t\t%d\n",symbolCount++, "Decode Sliceterm", bit);
//    fflush(p_trace);
//#endif
//  }
//  else
//  {
//    bit = 0;
//  }
//  
//  return (bit==1?1:0);
//}



/*!
 ************************************************************************
 * \brief
 *    Exp Golomb binarization and decoding of a symbol
 *    with prob. of 0.5
 ************************************************************************
 */
unsigned int exp_golomb_decode_eq_prob( DecodingEnvironment* dep_dp,
                                        int k)
{
  unsigned int l;
  int symbol = 0;
  int binary_symbol = 0;

  do
  {
    l=biari_decode_symbol_eq_prob(dep_dp);
    if (l==1) 
    {
      symbol += (1<<k); 
      k++;
    }
  }
  while (l!=0);

  while (k--)                             //next binary part
    if (biari_decode_symbol_eq_prob(dep_dp)==1) 
      binary_symbol |= (1<<k);

  return (unsigned int) (symbol+binary_symbol);
}


/*!
 ************************************************************************
 * \brief
 *    Exp-Golomb decoding for LEVELS
 ***********************************************************************
 */
unsigned int unary_exp_golomb_level_decode( DecodingEnvironment* dep_dp,
                                            BiContextType* ctx)
{
  unsigned int l,k;
  unsigned int symbol;
  unsigned int exp_start = 13;

  symbol = biari_decode_symbol(dep_dp, ctx );

  if (symbol==0)
    return 0;
  else
  {
    symbol=0;
    k=1;
    do
    {
      l=biari_decode_symbol(dep_dp, ctx);
      symbol++;
      k++;
    }
    while((l!=0) && (k!=exp_start));
    if (l!=0)
      symbol += exp_golomb_decode_eq_prob(dep_dp,0)+1;
    return symbol;
  }
}

/*!
 ************************************************************************
 * \brief
 *    Read one byte from CABAC-partition. 
 *    Bitstream->read_len will be modified
 *    (for IPCM CABAC  28/11/2003)
 *
 * \author
 *    Dong Wang <Dong.Wang@bristol.ac.uk>  
 ************************************************************************
*/
//void readIPCMBytes_CABAC(SyntaxElement *sym, Bitstream *currStream)
//{
//  int read_len = currStream->read_len;
//  int code_len = currStream->code_len;
//  byte *buf = currStream->streamBuffer;
//  
//  sym->len=8;
//  
//  if(read_len<code_len)
//    sym->inf=buf[read_len++];
//  
//  sym->value1=sym->inf;
//  
//  currStream->read_len=read_len;
//  
//#if TRACE
//  tracebits2(sym->tracestring, sym->len, sym->inf);
//#endif
//  
//}

