#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "alloc.h"
#include "codecdecl.h"



//
int zig_zag_scan(MacroBlock* mb, short int **img, int rows, int cols, int no_of_subbands)
{
	int i, j, k=0, pixel_cnt=0,l=0;
	int no_of_pixels;
	int old_val_x, old_val_y, new_val_x, new_val_y;
	int offset_x, offset_y; //offset from subband boundaries
	int no_hor_pixels, no_ver_pixels; //number of pixels pr. subband
	int direction_x, direction_y; //which direction the scan is moving {-1, 0, 1}
	int old_dir_x, old_dir_y;	  //old directions
	int no_skipped_mb = 0;		  //number of mb with none significant coeffs
	int last_significant_mb =0;

	no_of_pixels = rows*cols - rows*cols/(no_of_subbands*no_of_subbands);
	no_hor_pixels = cols/no_of_subbands;
	no_ver_pixels = rows/no_of_subbands;

	offset_x = offset_y = 0;		
	
	for (j=0; j<rows/no_of_subbands; j++) {
		for (i=0; i<cols/no_of_subbands; i++) {		
			old_val_x = i; //default values for each zig-zag scan
			old_val_y = j;
			direction_x=1;
			direction_y=0;
			pixel_cnt =0;
			//printf("\n\n\n");
			//default values
			mb[l].has_none_zero_coeffs=0; //equals no significant coeffs
			mb[l].position_x=i;
			mb[l].position_y=j;
			mb[l].bitcounter=0;
			mb[l].last_significant_coeff=0;
			mb[l].coeff_cnt=0;
			mb[l].mb_number = l;
			
			//check if there is available neighboor macroblocks
			if (i>0)
				mb[l].mb_available_left = &mb[l-1];
			else
				mb[l].mb_available_left = NULL;
			if (j>0)
				mb[l].mb_available_up = &mb[l-cols/no_of_subbands];
			else
				mb[l].mb_available_up = NULL;
					
			for (k=1;k<no_of_subbands*no_of_subbands; k++) {
				old_dir_x=direction_x;
				old_dir_y=direction_y;
				if (direction_x==1 && direction_y==0) { //go one-subband to the right
					new_val_x=old_val_x+no_hor_pixels;
					new_val_y=old_val_y;					
				}

				else if ((direction_x == -1) && (direction_y == 1)) {
					new_val_x=old_val_x-no_hor_pixels;
					new_val_y=old_val_y+no_ver_pixels;				
				}

				else if ((direction_x == 0) && (direction_y == 1)) {
					new_val_x=old_val_x;
					new_val_y=old_val_y+no_ver_pixels;
				}

				else if ((direction_x == 1) && (direction_y == -1)) {
					new_val_x=old_val_x+no_hor_pixels;
					new_val_y=old_val_y-no_ver_pixels;
				}

				else { //default
					break;
				}

				//get value, and add it to the MacroBlock
				//printf("%d \n", (int)img[new_val_y][new_val_x]);
				mb[l].coeffs[pixel_cnt] = img[new_val_y][new_val_x];
				if (img[new_val_y][new_val_x]!= 0) {
					mb[l].has_none_zero_coeffs=1;
					mb[l].last_significant_coeff=pixel_cnt;
					mb[l].coeff_cnt++;
				}
				pixel_cnt++;												

				//set new directions
				if ((new_val_x-i)==0) { 	//left border subband
					if (old_dir_x==0 && old_dir_y==1) { //last: downwards, now go diagonal
						direction_x = 1;
						direction_y = -1;							
					}										
					else if ((new_val_y+no_ver_pixels-j) >= (rows-1)) {//last subband in y-direction						
							  //go right
							direction_x=1;
							direction_y=0;						
					}
					else {					//not at last border subband in y-direction, go to subband below
						direction_x=0;
						direction_y=1;
					}
				}

				else if ((new_val_y-j)==0) {	//top y-subband										
					if (old_dir_x==1 && old_dir_y == 0) { //last: horizontal, now go diagonal down
						direction_x = -1;
						direction_y = 1;
					}
					
					else if ((new_val_x+no_hor_pixels-i) >= (cols-1)) { //last subband in x-direction, go down
						direction_x = 0;
						direction_y = 1;
					}
					
					else { //go horizontal to the right
						direction_x = 1;
						direction_y = 0;
					}					
				}

				else if ((new_val_y-j+no_ver_pixels)>=(rows-1)) {	//last subband in y-direction
					if (old_dir_x==0 && old_dir_y==1) { //last: went down, now go diagonal
						direction_x=1;
						direction_y=-1;
					}
					else if (old_dir_x==1 && old_dir_y==0) { //last: go right, now: go diagonal
						direction_x=1;
						direction_y=-1;							
					}		
					else {
						//go right
						direction_x = 1;
						direction_y = 0;
					}
				}

				else if ((new_val_x-i+no_hor_pixels)>=(cols-1)) {	//last subband in x-direction
					if (old_dir_x==1 && old_dir_y==0) { //last: went horizontal, now: go diagonal
						direction_x=-1;
						direction_y=1;
					}
					else if (old_dir_x==0 && old_dir_y==1) { //last: vertical, now: diagonal
						direction_x=-1;
						direction_y=1;
					}
					
					else {
						//go down
						direction_x=0;
						direction_y=1;
					}
				}
				
				else {
					//do not update the directions.
				}
				
				//update old_values
				old_val_x = new_val_x;
				old_val_y = new_val_y;
				
			} //end for(k=1....
			if (mb[l].has_none_zero_coeffs)
				last_significant_mb = l;
			else
				no_skipped_mb++;			
			l++;
		} //end for(i=0....
	} //end for(j=0....
	fprintf(stderr,"No of mb to be skipped= %d\n",no_skipped_mb);
	return last_significant_mb;
	//return one_dim_image;
}

void inv_zig_zag_scan(MacroBlock *mb,short int** img,int rows, int cols, int no_of_subbands)
{
	int i, j, k=0, pixel_cnt=0;
	//unsigned char **img;
	int no_of_pixels;
	int old_val_x, old_val_y, new_val_x, new_val_y;
	int offset_x, offset_y; //offset from subband boundaries
	int no_hor_pixels, no_ver_pixels; //number of pixels pr. subband
	int direction_x, direction_y; //which direction the scan is moving {-1, 0, 1}
	int old_dir_x, old_dir_y;	  //old directions

	no_of_pixels = rows*cols - rows*cols/(no_of_subbands*no_of_subbands);
	no_hor_pixels = cols/no_of_subbands;
	no_ver_pixels = rows/no_of_subbands;

	offset_x = offset_y = 0;
	direction_x = 1; //go right
	direction_y = 0; //dont't go in this direction
	old_val_x = old_val_y = 0;

	//img = matrix2_uch(rows, cols);

	
	for (j=0; j<rows/no_of_subbands; j++) {
		for (i=0; i<cols/no_of_subbands; i++) {		
			old_val_x = i; //default values for each zig-zag scan
			old_val_y = j;
			direction_x=1;
			direction_y=0;
//			printf("\n\n\n");
			for (k=0;k<no_of_subbands*no_of_subbands-1; k++) {
				old_dir_x=direction_x;
				old_dir_y=direction_y;
				if (direction_x==1 && direction_y==0) { //go one-subband to the right
					new_val_x=old_val_x+no_hor_pixels;
					new_val_y=old_val_y;					
				}

				else if ((direction_x == -1) && (direction_y == 1)) {
					new_val_x=old_val_x-no_hor_pixels;
					new_val_y=old_val_y+no_ver_pixels;				
				}

				else if ((direction_x == 0) && (direction_y == 1)) {
					new_val_x=old_val_x;
					new_val_y=old_val_y+no_ver_pixels;
				}

				else if ((direction_x == 1) && (direction_y == -1)) {
					new_val_x=old_val_x+no_hor_pixels;
					new_val_y=old_val_y-no_ver_pixels;
				}

				else { //default
					break;
				}

				//get value, and add it to the 1D-variable
//				printf("%d \n", (int)img[new_val_y][new_val_x]);
				img[new_val_y][new_val_x] = mb[pixel_cnt].coeffs[k];
//				img[new_val_y][new_val_x] = one_dim_image[pixel_cnt++];


				//set new directions
				if ((new_val_x-i)==0) { 	//left border subband
					if (old_dir_x==0 && old_dir_y==1) { //last: downwards, now go diagonal
						direction_x = 1;
						direction_y = -1;							
					}										
					else if ((new_val_y+no_ver_pixels-j) >= (rows-1)) {//last subband in y-direction						
							  //go right
							direction_x=1;
							direction_y=0;						
					}
					else {					//not at last border subband in y-direction, go to subband below
						direction_x=0;
						direction_y=1;
					}
				}

				else if ((new_val_y-j)==0) {	//top y-subband										
					if (old_dir_x==1 && old_dir_y == 0) { //last: horizontal, now go diagonal down
						direction_x = -1;
						direction_y = 1;
					}
					
					else if ((new_val_x+no_hor_pixels-i) >= (cols-1)) { //last subband in x-direction, go down
						direction_x = 0;
						direction_y = 1;
					}
					
					else { //go horizontal to the right
						direction_x = 1;
						direction_y = 0;
					}					
				}

				else if ((new_val_y-j+no_ver_pixels)>=(rows-1)) {	//last subband in y-direction
					if (old_dir_x==0 && old_dir_y==1) { //last: went down, now go diagonal
						direction_x=1;
						direction_y=-1;
					}
					else if (old_dir_x==1 && old_dir_y==0) { //last: go right, now: go diagonal
						direction_x=1;
						direction_y=-1;							
					}		
					else {
						//go right
						direction_x = 1;
						direction_y = 0;
					}
				}

				else if ((new_val_x-i+no_hor_pixels)>=(cols-1)) {	//last subband in x-direction
					if (old_dir_x==1 && old_dir_y==0) { //last: went horizontal, now: go diagonal
						direction_x=-1;
						direction_y=1;
					}
					else if (old_dir_x==0 && old_dir_y==1) { //last: vertical, now: diagonal
						direction_x=-1;
						direction_y=1;
					}
					
					else {
						//go down
						direction_x=0;
						direction_y=1;
					}
				}
				
				else {
					//do not update the directions.
				}
				
				//update old_values
				old_val_x = new_val_x;
				old_val_y = new_val_y;
				
			} //end for(k=1....
			pixel_cnt++; //next macroblock
		} //end for(i=0....
	} //end for(j=0....	
}
