#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include "alloc.h"
#include "EZW.h"
#include "list.h"
//#include "fifo.h"

int zeroes, ones;
//FILE *ezw_file;
unsigned char output_byte, mask;


void ezw_encode(float **Image,int RowsImage,int ColsImage, int NoOfSubbands,int levels,int frame,float LPQuant,float LPThreshold,unsigned char *Code,int *CodeCnt)
{
	float maxvalue;
	int BPP=0;
	EZWelement *EZW;
	float mean;	


	//init
	mean = 0.0;
	zeroes = 0; ones = 0;
	output_byte = 0;
	mask = 0x80;

	EZW = (EZWelement *)calloc(RowsImage*ColsImage/(NoOfSubbands*NoOfSubbands),sizeof(EZWelement));

	if (frame == 0) {
		//find mean
		mean = find_mean_and_subtract(Image,RowsImage,ColsImage,NoOfSubbands,levels,mean);

		memcpy(&Code[*CodeCnt],&mean,sizeof(float));
		*CodeCnt += sizeof(float);
//		fprintf(stderr,"mean = %f \n",mean);
		//subtract mean
		mean = find_mean_and_subtract(Image,RowsImage,ColsImage,NoOfSubbands,levels,mean);
	}

	maxvalue = max_value_in_matrix(Image,RowsImage,ColsImage,NoOfSubbands);	
	maxvalue -= LPThreshold;

	if (maxvalue<0) { //no values outside dead-zone
		Code[(*CodeCnt)++] = (unsigned char)0;
		Code[(*CodeCnt)++] = (unsigned char)0;
		return;
	}
	//find number of bits needed for the given maxvalue and LPquantstepsize

	
	BPP = ((int)(ceil(log10(maxvalue/LPQuant)/log10(2))));
	if (BPP <= 0){	//do one dominant pass only		
//		fprintf(stderr,"BPP =0, frame=%i\n",frame);
		BPP = 0;
		Code[(*CodeCnt)++] = (unsigned char)0;
		Code[(*CodeCnt)++] = (unsigned char)1;
	}
	else {				
		Code[(*CodeCnt)++] = (unsigned char)BPP;
	}
		
	EZW_code(EZW,Image,BPP,RowsImage,ColsImage,NoOfSubbands,levels,LPQuant,LPThreshold,Code,CodeCnt);

//	fprintf(stderr,"Mask= %i, output_byte= %i, CodeCnt= %i \n",(int)mask,(int)output_byte,*CodeCnt);

	if (mask!=0x80) {
    	memcpy(&Code[*CodeCnt],&output_byte,sizeof(char));
		*CodeCnt += sizeof(char);
  	}


  	destroy_fifo();
  	destroy_list();
	free(EZW);

}

void EZW_code(EZWelement *EZW,float **Image,int BPP,int RowsImage,int ColsImage, int NoOfSubbands,int levels,float LPQuant,float LPThreshold,unsigned char *Code,int *CodeCnt)
{
  int antall_codede = 0;
  float Threshold=LPThreshold;
  int tmp_bpp = BPP-1;
  int i;
  //Init_EZW
  init_EZW(Image,EZW,RowsImage,ColsImage,NoOfSubbands,levels);

  if (BPP==0) {//only do one dominant pass
  	tmp_bpp=0;
  	dominant_pass(EZW,RowsImage,ColsImage,NoOfSubbands,levels,Threshold,LPThreshold,Code,CodeCnt);
	destroy_list();
  	return;
  }    
  
  for (i=1;i<tmp_bpp;i++) {
	Threshold += (float)(1<<i)*LPQuant;
  }
  
  if (BPP==1)
  	tmp_bpp=1;
  //do coding
  while (tmp_bpp>0) {	
  	dominant_pass(EZW,RowsImage,ColsImage,NoOfSubbands,levels,Threshold,LPThreshold,Code,CodeCnt);
  	antall_codede += list_length;
  	subordinate_pass(Threshold,tmp_bpp,LPQuant,LPThreshold,Code,CodeCnt);
  	
  	tmp_bpp--;
	Threshold -= (float)(1<<tmp_bpp)*LPQuant;
  }
}

void dominant_pass(EZWelement *EZW,int RowsImage,int ColsImage,int NoOfSubbands,int levels,float threshold,float LPThreshold,unsigned char *Code,int *CodeCnt)
{
	int s;
	int tmp_number;
	int number=0;
	int i,j;
	int col_LP,row_LP;

	col_LP = (ColsImage/NoOfSubbands)>>levels;
	row_LP = (RowsImage/NoOfSubbands)>>levels;

	//all coeffs in the LP-LP subband (in the dyadic split)
	for (i=0; i<row_LP; i++) {
		for (j=0; j<col_LP; j++) {
			if (EZW[number].status == NC) {
				process_element(&EZW[number],threshold,LPThreshold);
				output_code(EZW[number].code,Code,CodeCnt);
			}
			put_in_fifo(number);
			number++;
		}
	}


	s = get_from_fifo();
	while (fifo_empty==0) {
    	if (EZW[s].code!=ZTR && EZW[s].code!=AZTR && s!=-1) {
    		if (EZW[s].up_left != NULL) {
    			if (EZW[s].up_left->status == NC) {
    				process_element(EZW[s].up_left,threshold,LPThreshold);
    				output_code(EZW[s].up_left->code,Code,CodeCnt);
    			}
    			tmp_number = coeff2scan(EZW[s].up_left->pos_x,EZW[s].up_left->pos_y,RowsImage,ColsImage,NoOfSubbands,levels);
    			put_in_fifo(tmp_number);
    		}
    		if (EZW[s].up_right != NULL) {
	    		if (EZW[s].up_right->status == NC) {
	    			process_element(EZW[s].up_right,threshold,LPThreshold);
	    			output_code(EZW[s].up_right->code,Code,CodeCnt);
	    		}
	    		tmp_number = coeff2scan(EZW[s].up_right->pos_x,EZW[s].up_right->pos_y,RowsImage,ColsImage,NoOfSubbands,levels);
    			put_in_fifo(tmp_number);
    		}
    		if (EZW[s].down_left != NULL) {
	    		if (EZW[s].down_left->status == NC) {
	    			process_element(EZW[s].down_left,threshold,LPThreshold);
	    			output_code(EZW[s].down_left->code,Code,CodeCnt);
	    		}
	    		tmp_number = coeff2scan(EZW[s].down_left->pos_x,EZW[s].down_left->pos_y,RowsImage,ColsImage,NoOfSubbands,levels);
    			put_in_fifo(tmp_number);
    		}
    		if (EZW[s].down_right != NULL) {
	    		if (EZW[s].down_right->status == NC) {
		    		process_element(EZW[s].down_right,threshold,LPThreshold);
		    		output_code(EZW[s].down_right->code,Code,CodeCnt);
	    		}
	    		tmp_number = coeff2scan(EZW[s].down_right->pos_x,EZW[s].down_right->pos_y,RowsImage,ColsImage,NoOfSubbands,levels);
    			put_in_fifo(tmp_number);
    		}
    	}
    	s = get_from_fifo();
	}
	destroy_fifo();
}

/*
 * Performs one subordinate pass.
 */
void subordinate_pass(float threshold,int BPP,float LPQuant,float LPThreshold,unsigned char *Code,int *CodeCnt)
{
  list_type d;
  int i,j;
  char found;
  int quantvalue;

  if (threshold>0) {
    for (i=0; i<list_length; i++) {
      d = get_list_element(i,&found);
      if (found==1) {
		//quantize each value
		quantvalue = (int)floor((d.x-threshold)/LPQuant);
		if (quantvalue >= (1<<BPP)) //example: quantvalue =512 when limit is 511
			quantvalue = (1<<BPP)-1;
//		fprintf(stderr,"value=%f and quantvalue=%i \n",d.x,quantvalue);

		//write the bits of the quantized value
		for (j=0;j<BPP;j++) {
			output_code((quantvalue>>j)&1,Code,CodeCnt);
		}
      }
    }
  }
  destroy_list(); //all coeffs in list has been coded with x bpp
}




void to_sub_list(float value)
{
  list_type d;
  /*
   * Put only coefficient magnitude in list, sign is allready coded.
   */
  d.x = fabs(value);
  append_to_list(d);
}


/*
 * Builds a dominant pass EZW-element from a matrix element and a threshold.
 */
void process_element(EZWelement *EZW,float threshold,float LPThreshold)
{
  EZW->code = code(EZW,threshold,LPThreshold);
//  fprintf(stderr,"x=%i, y=%i, code=%i \n",EZW->pos_x,EZW->pos_y,EZW->code);
  if ((EZW->code==POS) || (EZW->code==NEG)) {
//    fprintf(stderr,"x=%i , y=%i , value=%f \n",EZW->pos_x,EZW->pos_y,EZW->value);
    to_sub_list(EZW->value);
    EZW->value = 0;
  }
}



/*
 * Returns a dominant-pass code from the alphabet [POS,NEG,ZTR,IZ].
 */
int code(EZWelement *EZW,float threshold,float LPThreshold)
{
  float temp;
  float max = 0.0;
  float max_des;
  temp = EZW->value;
  if (fabs(temp)>=threshold) {
    EZW->status = 1; //CODED
    if (temp>=0) return (POS);
    else return (NEG);
  }
  else {
    max_des=max_descendant(EZW,max);
    if (max_des<LPThreshold) {
    	//set element and all children to status=coded
    	set_state(EZW,1);
    	return (AZTR);
    }
    else if (max_des<threshold)
    	return (ZTR);
    else return (IZ);
  }
}



void init_EZW(float **Image,EZWelement *EZW,int RowsImage,int ColsImage, int NoOfSubbands, int levels)
{
	int x,y;
	int number;
	int parent_number;
	for (y=0; y<RowsImage/NoOfSubbands; y++) {
		for (x=0; x<ColsImage/NoOfSubbands; x++) {
			number = coeff2scan(x,y,RowsImage,ColsImage,NoOfSubbands,levels);
			EZW[number].pos_x = x;
			EZW[number].pos_y = y;

			//set parent
			parent_number = get_parent(number,x,y,RowsImage,ColsImage,NoOfSubbands,levels);
			if (parent_number != -1)
				EZW[number].parent = &EZW[parent_number];
			else
				EZW[number].parent = NULL;

			//set children
			get_children(EZW,number,x,y,RowsImage,ColsImage,NoOfSubbands,levels);



			EZW[number].value = Image[y][x];
			EZW[number].status = NC; //not coded
		}
	}
}

void get_children(EZWelement *EZW,int number,int x,int y,int RowsImage,int ColsImage,int NoOfSubbands,int levels) {
	int child_nr;
	int child_x,child_y;
	int coeffs_in_LP=(RowsImage*ColsImage/(NoOfSubbands*NoOfSubbands))>>(2*levels);
	int col_level1,row_level1,col_level2,row_level2;

	col_level1 = (ColsImage/NoOfSubbands)>>1;
	row_level1 = (RowsImage/NoOfSubbands)>>1;
	col_level2 = col_level1>>1;
	row_level2 = row_level1>>1;

	if (levels ==3)	 {

	}
	else if (levels == 2) {
		//in LL2
		if (number < coeffs_in_LP) {
			//only 3 children
			EZW[number].up_left = NULL;

			//up right
			child_x = x+col_level2;
			child_y = y;
			child_nr = coeff2scan(child_x,child_y,RowsImage,ColsImage,NoOfSubbands,levels);
			EZW[number].up_right = &EZW[child_nr];

			//down left
			child_x = x;
			child_y = y+row_level2;
			child_nr = coeff2scan(child_x,child_y,RowsImage,ColsImage,NoOfSubbands,levels);
			EZW[number].down_left = &EZW[child_nr];

			//down right
			child_x = x+col_level2;
			child_y = y+row_level2;
			child_nr = coeff2scan(child_x,child_y,RowsImage,ColsImage,NoOfSubbands,levels);
			EZW[number].down_right = &EZW[child_nr];
		}

		//4 children

		//in HL2
		else if (number < 2*coeffs_in_LP) {
			child_x = col_level1 + (x-col_level2)*2;
			child_y = y*2;
			child_nr = coeff2scan(child_x,child_y,RowsImage,ColsImage,NoOfSubbands,levels);
			EZW[number].up_left = &EZW[child_nr];
			EZW[number].up_right = &EZW[child_nr+1];
			EZW[number].down_left = &EZW[child_nr+2];
			EZW[number].down_right = &EZW[child_nr+3];
		}
		//in LH2
		else if (number < 3*coeffs_in_LP) {
			child_x = 2*x;
			child_y = row_level1 + (y-row_level2)*2;
			child_nr = coeff2scan(child_x,child_y,RowsImage,ColsImage,NoOfSubbands,levels);
			EZW[number].up_left = &EZW[child_nr];
			EZW[number].up_right = &EZW[child_nr+1];
			EZW[number].down_left = &EZW[child_nr+2];
			EZW[number].down_right = &EZW[child_nr+3];
		}
		else if (number < 4*coeffs_in_LP) {
			child_x = col_level1 + (x-col_level2)*2;
			child_y = row_level1 + (y-row_level2)*2;
			child_nr = coeff2scan(child_x,child_y,RowsImage,ColsImage,NoOfSubbands,levels);
			EZW[number].up_left = &EZW[child_nr];
			EZW[number].up_right = &EZW[child_nr+1];
			EZW[number].down_left = &EZW[child_nr+2];
			EZW[number].down_right = &EZW[child_nr+3];
		}
		//in HL1, LH1 or HH1  - no children
		else {
			EZW[number].up_left = NULL;
			EZW[number].up_right = NULL;
			EZW[number].down_left = NULL;
			EZW[number].down_right = NULL;
		}
	}
	else {

	}
}


int get_parent(int number,int x,int y,int RowsImage,int ColsImage,int NoOfSubbands,int levels)
{
	int parent_nr;
	int coeffs_in_LP=(RowsImage*ColsImage/(NoOfSubbands*NoOfSubbands))>>(2*levels);
	int col_level1,row_level1,col_level2,row_level2;

	col_level1 = (ColsImage/NoOfSubbands)/2;
	row_level1 = (RowsImage/NoOfSubbands)/2;
	col_level2 = col_level1/2;
	row_level2 = row_level1/2;


	//in LP-LP band, no parent
	if (number < coeffs_in_LP) {
		return -1; //no parent
	}

	//in same dyadic split level
	else if (number < 4*coeffs_in_LP) {
		if (x >= col_level2)
			x-= col_level2;
		if (y >= row_level2)
			y -= row_level2;
		parent_nr = coeff2scan(x,y,RowsImage,ColsImage,NoOfSubbands,levels);
	}
	else {
		coeffs_in_LP <<=2;
		x = x/2;
		y = y/2;
		parent_nr = coeff2scan(x,y,RowsImage,ColsImage,NoOfSubbands,levels);
	}
	return parent_nr;
}

void set_state(EZWelement *EZW,unsigned char state) {
	EZW->status = state;
	if (EZW->up_left != NULL)
		set_state(EZW->up_left,state);
	if (EZW->up_right != NULL)
		set_state(EZW->up_right,state);
	if (EZW->down_left != NULL)
		set_state(EZW->down_left,state);
	if (EZW->down_right != NULL)
		set_state(EZW->down_right,state);
}


// from x,y to scan_number
int coeff2scan(int x, int y, int RowsImage, int ColsImage, int NoOfSubbands, int levels) {
	int number=0;
	int coeff_pr_parent;
	int pix_band3, pix_band2,pix_band1;
	int coeff_in_block_col;
	int pix_LPLP = RowsImage*ColsImage/(NoOfSubbands*NoOfSubbands);
	int cols_pr_band_level_one,rows_pr_band_level_one;
	int cols_pr_band_level_two,rows_pr_band_level_two;

	pix_band1 = pix_LPLP/4;

	cols_pr_band_level_one = (ColsImage/NoOfSubbands)/2;
	rows_pr_band_level_one = (RowsImage/NoOfSubbands)/2;

	cols_pr_band_level_two = cols_pr_band_level_one/2;
	rows_pr_band_level_two = rows_pr_band_level_one/2;

	if (levels == 3) {
		fprintf(stderr,"coeff2scan not implementet for level=3");
		exit(1);

	}
	else if (levels == 2) {
		pix_band2 = pix_band1/4;
		//in dyadic split region
		if (x<cols_pr_band_level_one && y<rows_pr_band_level_one) {
			//in LL2
			if (x<cols_pr_band_level_two && y<rows_pr_band_level_two) {
				number = y*cols_pr_band_level_two + x;
			}
			//in HL2
			else if (x >= cols_pr_band_level_two && y<rows_pr_band_level_two) {
				number = pix_band2;
				x -= cols_pr_band_level_two;
				number += y*cols_pr_band_level_two + x;
			}
			//in LH2
			else if (x<cols_pr_band_level_two && y >= rows_pr_band_level_two) {
				number = 2*pix_band2;
				y -= rows_pr_band_level_two;
				number += y*cols_pr_band_level_two + x;
			}
			//in HH2
			else {
				number = 3*pix_band2;
				x -= cols_pr_band_level_two;
				y -= rows_pr_band_level_two;
				number += y*cols_pr_band_level_two + x;
			}
			return number;
		}
		//in HL1,LH1 or HH1
		else {
			coeff_pr_parent = 4;
			coeff_in_block_col = cols_pr_band_level_one/2*coeff_pr_parent;
			//in HL1
			if (x >= cols_pr_band_level_one && y<rows_pr_band_level_one) {
				number = pix_band1;
				x -= cols_pr_band_level_one;
				number += (y/2)*coeff_in_block_col + (x/2)*coeff_pr_parent + (y%2)*2 + (x%2);
			}
			//in LH1
			else if (x<cols_pr_band_level_one && y >= rows_pr_band_level_one) {
				number = 2*pix_band1;
				y -= rows_pr_band_level_one;
				number += (y/2)*coeff_in_block_col + (x/2)*coeff_pr_parent + (y%2)*2 + (x%2);
			}
			//in HH1
			else {
				number = 3*pix_band1;
				y -= rows_pr_band_level_one;
				x -= cols_pr_band_level_one;
				number += (y/2)*coeff_in_block_col + (x/2)*coeff_pr_parent + (y%2)*2 + (x%2);
			}
			return number;
		}
	}
	else { //level = 1
		pix_band3 = pix_band2/4;
		fprintf(stderr,"coeff2scan not implementet for level=1");
		exit(1);
	}
	return number;



}


void scan2coeff(int number,int *x,int *y, int RowsImage,int ColsImage,int NoOfSubbands,int levels) {
	int pix_band3, pix_band2,pix_band1;
	int rows_pr_band,cols_pr_band;
	int start_x,start_y;
	int coeff_pr_parent;
	int block_nr,blocks_pr_col;
	int pix_LPLP = RowsImage*ColsImage/(NoOfSubbands*NoOfSubbands);

	pix_band1 = pix_LPLP/4;
	if (levels>1)
		pix_band2 = pix_band1/4;
	else
		pix_band2 = 0;
	if (levels>2)
		pix_band3 = pix_band2/4;
	else
		pix_band3 = 0;

	if (levels == 3) {
		fprintf(stderr,"scan2coeff not implemented for level=3");
		exit(1);



	}

	else if (levels == 2) {
		rows_pr_band = RowsImage>>levels;
		cols_pr_band = ColsImage>>levels;

		//in LL2
		if (number<pix_band2) {
			*x = number%cols_pr_band;
			*y = number/cols_pr_band;
		}
		//in HL2
		else if (number<2*pix_band2) {
			*x = cols_pr_band + (number-pix_band2)%cols_pr_band;
			*y = (number-pix_band2)/cols_pr_band;
		}
		//in LH2
		else if (number<3*pix_band2) {
			*x = (number-2*pix_band2)%cols_pr_band;
			*y = rows_pr_band + (number-2*pix_band2)/cols_pr_band;
		}
		//in HH2
		else if (number<pix_band1) {
			*x = cols_pr_band + (number-3*pix_band2)%cols_pr_band;
			*y = rows_pr_band + (number-3*pix_band2)/cols_pr_band;
		}

		else {
			coeff_pr_parent = 4;
			rows_pr_band = RowsImage>>(levels-1);
			cols_pr_band = ColsImage>>(levels-1);

			//blocks_pr_col = bredde HL1 /2 (to coeff pr blokk i bredden)
			blocks_pr_col = (ColsImage/NoOfSubbands)/4;
			//in HL1
			if (number<2*pix_band1) {

				start_x = (ColsImage/NoOfSubbands)/2;
				start_y = 0;
				number -= pix_band1;

				//START HER!!!!!!
				block_nr = number/coeff_pr_parent;
				start_x += block_nr%blocks_pr_col;
				start_y += 2*block_nr/blocks_pr_col;

				number -= coeff_pr_parent*(number/coeff_pr_parent);
				if (number == 1)
					start_x++;
				else if (number == 2)
					start_y++;
				else if (number == 3) {
					start_x++;
					start_y++;
				}
				*x = start_x;
				*y = start_y;
			}
			//in LH1
			else if (number<3*pix_band1) {
				start_x = 0;
				start_y = (RowsImage/NoOfSubbands)/2;
				number -= 2*pix_band1;

				block_nr = number/coeff_pr_parent;
				start_x += block_nr%blocks_pr_col;
				start_y += 2*block_nr/blocks_pr_col;

				number -= coeff_pr_parent*(number/coeff_pr_parent);
				if (number == 1)
					start_x++;
				else if (number == 2)
					start_y++;
				else if (number == 3) {
					start_x++;
					start_y++;
				}
				*x = start_x;
				*y = start_y;
			}
			//in HH1
			else {
				start_x = (ColsImage/NoOfSubbands)/2;
				start_y = (RowsImage/NoOfSubbands)/2;
				number -= 3*pix_band1;

				block_nr = number/coeff_pr_parent;
				start_x += block_nr%blocks_pr_col;
				start_y += 2*block_nr/blocks_pr_col;

				number -= coeff_pr_parent*(number/coeff_pr_parent);
				if (number == 1)
					start_x++;
				else if (number == 2)
					start_y++;
				else if (number == 3) {
					start_x++;
					start_y++;
				}
				*x = start_x;
				*y = start_y;
			}
		}
	}

	else {
		fprintf(stderr,"scan2coeff not implemented for level=1");
		exit(1);
	}
}


/*
 * Puts dominant-pass and subordinate-pass codes in the output stream.
 */
void output_code(int code,unsigned char *Code,int *CodeCnt)
{
  switch (code) {
    case ZERO:
      put_bit('0',Code,CodeCnt);
    break;

    case ONE:
      put_bit('1',Code,CodeCnt);
    break;

    case POS:
      put_bit('1',Code,CodeCnt);
      put_bit('0',Code,CodeCnt);
    break;

    case NEG:
      put_bit('1',Code,CodeCnt);
      put_bit('1',Code,CodeCnt);
    break;

    case ZTR:
      put_bit('0',Code,CodeCnt);
      put_bit('0',Code,CodeCnt);
    break;

    case IZ:
      put_bit('0',Code,CodeCnt);
      put_bit('1',Code,CodeCnt);
      put_bit('0',Code,CodeCnt);
    break;

    case AZTR:
      put_bit('0',Code,CodeCnt);
      put_bit('1',Code,CodeCnt);
      put_bit('1',Code,CodeCnt);
    break;
  }
}

/*
 * Puts a bit in the output stream.
 */
void put_bit(char bit,unsigned char *Code,int *CodeCnt)
{
  if (bit=='1') {
    output_byte |= mask;
    ones++;
  }
  else zeroes++;

  mask >>= 1;
  if (mask==0) {
    Code[(*CodeCnt)++] = output_byte;
    output_byte = 0;
    mask = 0x80;
  }
}

void ReadFrameFromFile(char txtfile[80],float **Image,int RowsImage,int ColsImage) {
	FILE *ftxt;
	int i,j;
	if ((ftxt = fopen(txtfile,"rb")) == NULL){
    	fprintf(stderr,"Cannot open file %s for reading\n",txtfile);
	    //exit(1);
	    return;
  	}
  	for (i=0;i<RowsImage;i++) {
  		j = fread(Image[i],sizeof(float),ColsImage,ftxt);
  		if (j != ColsImage) {
  			fprintf(stderr,"Could not read from file, j=%i \n",j);
  			j=fclose(ftxt);
  			return;
  		}
  	}
  	i = fclose(ftxt);
  	if (i != 0) {
  		fprintf(stderr, "fclose returned %i \n",i);
  	}
}

float max_value_in_matrix(float **Image,int RowsImage,int ColsImage,int NoOfSubbands) {
	int i,j;
	int y,x;
	float max;
	max = 0.0;

	for (i=0;i<RowsImage/NoOfSubbands; i++) {
		for (j=0; j<ColsImage/NoOfSubbands; j++) {
			if (abs(Image[i][j])>max) {
				max = abs(Image[i][j]);
				x=j;
				y=i;
			}
		}
	}
//	fprintf(stderr,"maxvalue in matrix=%f , x=%i , y=%i \n",max,x,y);
	return max;
}

float find_mean_and_subtract(float **Image,int RowsImage,int ColsImage,int NoOfSubbands,int levels,float subvalue)
{
	int row_LP, col_LP;
	float mean=0.0;
	int i,j;

	col_LP = (ColsImage/NoOfSubbands)>>levels;
	row_LP = (RowsImage/NoOfSubbands)>>levels;

	for (i=0;i<row_LP;i++) {
		for (j=0;j<col_LP;j++) {
			Image[i][j] -= subvalue;
			mean += Image[i][j];
		}
	}
	return (mean/(col_LP*row_LP));
}

float max_descendant(EZWelement *EZW,float max) {
	//don't check value on parent node

	if (fabs(EZW->value) > max)
		max = fabs(EZW->value);
	//get max value from all descendant-nodes
	if (EZW->up_left != NULL)
		max = max_descendant(EZW->up_left,max);
	if (EZW->up_right != NULL)
		max = max_descendant(EZW->up_right,max);
	if (EZW->down_left != NULL)
		max = max_descendant(EZW->down_left,max);
	if (EZW->down_right != NULL)
		max = max_descendant(EZW->down_right,max);

	return max;
}
