#include <stdio.h>
#include <time.h>
#include "codecdecl.h"
#include "util.h"
#include "alloc.h"


int ImageEncode(float **ImageArray,unsigned char **Code,struct EncoderArgs *Args,FILE *file, int WriteAllParameters, int CodeCnt,CABAC_EncoderEnvironment* CodingEnvironment,int *EZWSize,double *CABAC_time)
{  /* BEGIN MAIN */
	short int **QuantIndexArrayY;
	int last_significant_mb;
	clock_t start_clock,stop_clock;

	int i, j,k,m;
	int RowsImage, ColsImage, FileSize,buffer;
	float LPQuant,LPThreshold;
	int NoOfSubbands,levels;
	int Rows_old, Cols_old;
	char TmpName[40];
	float QuantStepSize, Threshold;
	float **SubbandImageY;
	MacroBlock *MB;


	//CABAC variables
	int model_number, img_type, ctx_number;
	EncodingEnvironment* eep;
	MotionInfoContexts* mot_ctx;
	TextureInfoContexts* tex_ctx;


	/* Initialization */
	RowsImage=Args->height;
	ColsImage=Args->width;
	Rows_old = Args->height_old;
	Cols_old = Args->width_old;
	buffer = Args->buffer;
	levels = (int)Args->levels;
	QuantStepSize=(float)Args->quality;
	Threshold = (float)Args->threshold;
	LPQuant = Args->LPquant;
	LPThreshold = Args->LPthreshold;
	CodeCnt = 0;
	NoOfSubbands=NOOFSUBBANDSLUMIN;



	/* Encoder parameters */
	*Code=matrix1_uch(MAXCODELENGTH);

	//CABAC
	eep = &(CodingEnvironment->partArr)->ee_cabac;
	img_type = 2;
	ctx_number = 0;
	mot_ctx = CodingEnvironment->mot_ctx; //ContextsModels
	tex_ctx = CodingEnvironment->tex_ctx;


	if (WriteAllParameters ==1 ) {
		write_All_Encoder_Parameters(Args,*Code, &CodeCnt);
		//fprintf(stderr, "After writing encoder parameters, codesize=%i \n",CodeCnt);
	}


	//allocate memory for the Macroblocks
	MB = (MacroBlock *)calloc(RowsImage*ColsImage/(NoOfSubbands*NoOfSubbands),sizeof(MacroBlock));
	QuantIndexArrayY=matrix2_si(RowsImage,ColsImage);
	SubbandImageY=matrix2_f(RowsImage,ColsImage);

//***************************************************************


	for (m=0; m<buffer; m++) {
		//read one frame
		k=0;
		for(i=0;i<RowsImage;i++) {
			for (j=0; j<ColsImage;j++) {
				SubbandImageY[i][j] = ImageArray[k++][m];
			}
		}

//***************************************************************

		/* Uniform quantization and EZW coding of LP-LP band */
		/*fprintf(stderr,"Uniform quantization...\n");*/
//		sprintf(TmpName,"LPenc%i.txt",m);
//		writeTxtFile2_f(TmpName, SubbandImageY, RowsImage/8, ColsImage/8);



		//EZW of LP-LP band
		ezw_encode(SubbandImageY,RowsImage,ColsImage,NoOfSubbands,levels,m,LPQuant,LPThreshold,*Code,&CodeCnt);
		*EZWSize += CodeCnt;


		unifquant(SubbandImageY,QuantIndexArrayY,RowsImage,ColsImage,NoOfSubbands,QuantStepSize,Threshold,*Code,&CodeCnt,m,buffer,levels);

//		writeTxtFile2_uch("EncodeLPLPQuantIndexArray.txt", QuantIndexArrayY, RowsImage/8, ColsImage/8);
//		sprintf(TmpName,"QuantIndexArray%i.txt",m);
//		writeTxtFile2_si(TmpName, QuantIndexArrayY, RowsImage, ColsImage);


//***************************************************************

		//ZIG-ZAG scan
		last_significant_mb = zig_zag_scan(MB, QuantIndexArrayY,RowsImage,ColsImage,NoOfSubbands);

		start_clock = clock();
		//initialize arithmetic environment for this frame
		arienco_start_encoding(eep,*Code,&CodeCnt);
		model_number = SetCtxModelNumber(0,img_type,ctx_number);
		init_contexts(mot_ctx,tex_ctx,model_number,m,buffer,QuantStepSize);
		//encode the blocks
		write_macroblock(MB,CodingEnvironment,RowsImage,ColsImage,NoOfSubbands,last_significant_mb);
		//terminate encoder for this frame
		write_terminating_bit(1,CodingEnvironment);
		arienco_done_encoding(eep);
		stop_clock = clock();
		*CABAC_time += (double)(stop_clock-start_clock);





//***************************************************************

		//write to CODEFILE
		FileSize = fwrite(*Code,sizeof(unsigned char),CodeCnt,file);
		if (FileSize != CodeCnt) {
			fprintf(stderr, "Error writing to CODEFILE, Codesize=%i written size=%i\n",CodeCnt,FileSize);
		}
//		fprintf(stderr,"ftell(file) = %lu \n",ftell(file));
		Args->codesize+=CodeCnt;
		CodeCnt=0;

		//set all data in MB to zero
		MB = (MacroBlock*)memset(MB,0,RowsImage*ColsImage/(NoOfSubbands*NoOfSubbands)*sizeof(MacroBlock));
	}

	//free up memory
	free(MB);
	free(*Code);
	freematrix2_f(SubbandImageY, RowsImage, ColsImage);
	freematrix2_si(QuantIndexArrayY,RowsImage,ColsImage);
	return(0);
}

//writes all Encoder Parameters
void write_All_Encoder_Parameters(struct EncoderArgs *Args,unsigned char *Code,int *CodeCnt) {
	unsigned char tmpBuffer;

	memcpy(&Code[*CodeCnt],&Args->height,sizeof(int));
	*CodeCnt+=sizeof(int);
	memcpy(&Code[*CodeCnt],&Args->width,sizeof(int));
	*CodeCnt+=sizeof(int);
	memcpy(&Code[*CodeCnt],&Args->quality,sizeof(float));
	*CodeCnt+=sizeof(float);
	memcpy(&Code[*CodeCnt],&Args->threshold,sizeof(float));
	*CodeCnt+=sizeof(float);
	memcpy(&Code[*CodeCnt],&Args->height_old,sizeof(int));
	*CodeCnt+=sizeof(int);
	memcpy(&Code[*CodeCnt],&Args->width_old,sizeof(int));
	*CodeCnt+=sizeof(int);
	tmpBuffer = (unsigned char)Args->buffer;
	memcpy(&Code[*CodeCnt],&Args->TotNoOFFrames,sizeof(int));
	*CodeCnt+=sizeof(int);
	memcpy(&Code[*CodeCnt],&tmpBuffer,sizeof(unsigned char));
	*CodeCnt+=sizeof(char);
	memcpy(&Code[*CodeCnt],&Args->levels,sizeof(unsigned char));
	*CodeCnt+=sizeof(char);
	memcpy(&Code[*CodeCnt],&Args->LPquant,sizeof(float));
	*CodeCnt+=sizeof(float);
	memcpy(&Code[*CodeCnt],&Args->LPthreshold,sizeof(float));
	*CodeCnt+=sizeof(float);
}
