/*!
  \file
  \brief Spatial Decomposition:  8 band parallel + 3 dyadic splits.
*/

/***************************************************************************
* Spatial Decomposition:  8 band parallel + 3 dyadic splits
*--------------------------------------------------------------------------
* Input:      original image (raw format)
* Output:     subband image (sub format)
* Parameters: number of rows
*             number of columns
*             bit depth
*--------------------------------------------------------------------------
* Original version        Arild Fuldseth
* This version            Helge Coward          2001-03-23
***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "constants.h"
#include "filt_par.h"
#include "filt_tree.h"
#include "filtering_k.h"
#include "util.h"
#include "alloc.h"


int analysis(float **filterdct, int nrows,int ncols,float **filt_input,int maxlevels,int buffer,int nrows_exp, int ncols_exp) {
	int nrowse,ncolse,i,j;
	int div2,mindim,screenflag=0;
	//int v_blksz=8, h_blksz=8;
	float          *x,*xe,*sub1,*sub2;



	if (nrows > MAXDIM || ncols > MAXDIM || nrows*ncols > MAXSIZE ){
		fprintf(stderr,"Image too large\n");
		exit(-1);
	}

	if (screenflag) {
		fprintf(stderr,"\nANALYSIS_K\n\n");
	}

	mindim = (nrows < ncols) ? nrows : ncols;

	if (mindim < (128-15) || maxlevels < 1){
		fil_tree.levels = 0;
	}
	else if (mindim < (256-31) || maxlevels <2){
		fil_tree.levels = 1;
	}
	else if (mindim < (512-63) || maxlevels < 3){
		fil_tree.levels = 2;
	}
	else{
		fil_tree.levels = 3;
	}

	div2 = 1<<(NPARLEVELS+fil_tree.levels);

//	expanded_size(nrows, ncols, v_blksz, h_blksz, NPARLEVELS,
//					 fil_tree.levels, &nrowse, &ncolse);
	nrowse =nrows_exp;
	ncolse = ncols_exp;
  
	for (i=0;i<buffer;i++) {
		x = matrix1_f(ncols*nrows);
		for(j=0;j<nrows*ncols;j++) {
			x[j] = filt_input[j][i];
		}
	  
		/* Extend the image */
		if ((nrowse!=nrows) || (ncolse!=ncols)) {
			xe = matrix1_f(nrowse*ncolse);
			Extend_Image(nrows,ncols,nrowse,ncolse,x,xe,screenflag);
			free(x);
		}
		else {
			xe = x;
		}
	
		/* filter image */	
		sub1 = matrix1_f(nrowse*ncolse);
		Anl_fbank_Par(nrowse,ncolse,xe,sub1,fil_par,screenflag);
		free(xe);
		if (fil_tree.levels > 0){
			sub2 = matrix1_f(nrowse*ncolse);
			Anl_fbank_Tree(nrowse,ncolse,sub1,sub2,fil_tree,screenflag);
			free(sub1);
		}
		else{
			sub2=sub1;
		}
	  
		for(j=0; j<nrowse*ncolse; j++) {
			filterdct[j][i] = sub2[j];
		}
		free(sub2);  
	}
	return fil_tree.levels;
}
