#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "biariencode.h"
#include "codecdecl.h"


int binCount = 0;
int cabac_encoding;

//
//#define Elow                      (eep->Elow)
//#define Erange                    (eep->Erange)
//#define Ebits_to_follow           (eep->Ebits_to_follow)
//#define Ebuffer                   (eep->Ebuffer)
//#define Ebits_to_go               (eep->Ebits_to_go)
//#define Ecodestrm                 (eep->Ecodestrm)
//#define Ecodestrm_len             (eep->Ecodestrm_len)
//#define Ecodestrm_laststartcode   (eep->Ecodestrm_laststartcode)
/*!
 ************************************************************************
 * Macro for writing bytes of code
 ***********************************************************************
 */

void put_byte(EncodingEnvironment* eep) {	
	eep->Ecodestrm[(*eep->Ecodestrm_len)++] = eep->Ebuffer;
	eep->Ebits_to_go = 8;
	while (eep->C > 7) {
		eep->C-=8;
		eep->E++;
	}
}
													//eep->Ebuffer*2 + bit(b)
void put_one_bit(EncodingEnvironment* eep,byte b) {	//eep->Ebuffer = eep->Ebuffer<<1
	eep->Ebuffer <<= 1; eep->Ebuffer |= (b); 
	if (--(eep->Ebits_to_go) == 0) 
		put_byte(eep); 
}

void put_one_bit_plus_outstanding(EncodingEnvironment* eep,byte b) { 
	put_one_bit(eep,b); 
	while (eep->Ebits_to_follow > 0) 
	{ 
		eep->Ebits_to_follow--; 
		put_one_bit(eep,!(b)); 
	} 
}

int pic_bin_count;

void reset_pic_bin_count()
{
  pic_bin_count = 0;
}

int get_pic_bin_count()
{
  return pic_bin_count;
}

/*!
 ************************************************************************
 * \brief
 *    Allocates memory for the EncodingEnvironment struct
 ************************************************************************
 */
EncodingEnvironment* arienco_create_encoding_environment()
{
  EncodingEnvironment* eep;

  if ( (eep = (EncodingEnvironment*)calloc(1,sizeof(EncodingEnvironment))) == NULL) {
    //no_mem_exit("arienco_create_encoding_environment: eep");
    fprintf(stderr,"error, arienvo_create_encoding_environment");
    exit(1);
  }

  return eep;
}

/*!
 ************************************************************************
 * \brief
 *    Frees memory of the EncodingEnvironment struct
 ************************************************************************
 */
void arienco_delete_encoding_environment(EncodingEnvironment* eep)
{
  if (eep == NULL)
  {
    fprintf(stderr, "Error freeing eep (NULL pointer)");
    //error (errortext, 200);
  }
  else
    free(eep);
}


/*!
 ************************************************************************
 * \brief
 *    Initializes the EncodingEnvironment for the arithmetic coder
 ************************************************************************
 */
void arienco_start_encoding(EncodingEnvironment* eep,
                            unsigned char *code_buffer,
                            int *code_len )
{
  eep->Elow = 0;
  eep->Ebits_to_follow = 0;
  eep->Ebuffer = 0;
  eep->Ebits_to_go = 9; // to swallow first redundant bit

  eep->Ecodestrm = code_buffer;
  eep->Ecodestrm_len = code_len;

  eep->Erange = HALF-2;

  eep->C = 0;
  eep->E = 0;
}

/*!
 ************************************************************************
 * \brief
 *    Returns the number of currently written bits
 ************************************************************************
 */
int arienco_bits_written(EncodingEnvironment* eep)
{
   return (8 * (*eep->Ecodestrm_len) + eep->Ebits_to_follow + 8  - eep->Ebits_to_go);
}

/*!
 ************************************************************************
 * \brief
 *    Terminates the arithmetic codeword, writes stop bit and stuffing bytes (if any)
 ************************************************************************
 */
void arienco_done_encoding(EncodingEnvironment* eep)
{
  put_one_bit_plus_outstanding(eep, (unsigned char) ((eep->Elow >> (B_BITS-1)) & 1));
  put_one_bit(eep,(unsigned char) (eep->Elow >> (B_BITS-2))&1);
  put_one_bit(eep,(unsigned char) 1);

  //stats->bit_use_stuffingBits[img->type]+=(8-Ebits_to_go);

  while (eep->Ebits_to_go != 8)
    put_one_bit(eep,0);

  pic_bin_count += eep->E*8 + eep->C; // no of processed bins
}



void biari_encode_symbol(EncodingEnvironment* eep, signed short symbol, BiContextType* bi_ct )
{
  register unsigned int range = eep->Erange;
  register unsigned int low = eep->Elow;  
  unsigned int rLPS = rLPS_table_64x4[bi_ct->state][(range>>6) & 3];
//
//#if (2==TRACE)
//  if (cabac_encoding)
//    fprintf(p_trace, "%d  0x%04x  %d  %d\n", binCount++, Erange , bi_ct->state, bi_ct->MPS );
//#endif
//
  range -= rLPS;
  bi_ct->count += cabac_encoding;

  /* covers all cases where code does not bother to shift down symbol to be
   * either 0 or 1, e.g. in some cases for cbp, mb_Type etc the code simply
   * masks off the bit position and passes in the resulting value */
  symbol = (short) (symbol != 0);

  if (symbol != bi_ct->MPS)
  {
    low += range;
    range = rLPS;

    if (!bi_ct->state)
      bi_ct->MPS = (unsigned char) (bi_ct->MPS ^ 0x01);               // switch LPS if necessary
    bi_ct->state = AC_next_state_LPS_64[bi_ct->state]; // next state

    //AC_next_state_LPS_64 og AC_next_state_MPS_64 definert i biariencode.h
  }
  else
    bi_ct->state = AC_next_state_MPS_64[bi_ct->state]; // next state

  /* renormalisation */
  while (range < QUARTER)						//QUARTER = 1<<8 = 2^8 = 256
  {												//HALF = 1<<9 = 512
    if (low >= HALF)							//ONE = 1<<10 = 1024
    {
      put_one_bit_plus_outstanding(eep, 1);
      low -= HALF;
    }
    else if (low < QUARTER)
    {
      put_one_bit_plus_outstanding(eep, 0);
    }
    else
    {
      eep->Ebits_to_follow++;
      low -= QUARTER;
    }
    low <<= 1;
    range <<= 1;		//range = range<<1 (range *2)
  }
  eep->Erange = range;
  eep->Elow = low;
  eep->C++;
}

/*!
 ************************************************************************
 * \brief
 *    Arithmetic encoding of one binary symbol assuming
 *    a fixed prob. distribution with p(symbol) = 0.5
 ************************************************************************
 */
void biari_encode_symbol_eq_prob(EncodingEnvironment* eep, signed short symbol)
{
  register unsigned int low = (eep->Elow<<1);
//
//#if (2==TRACE)
//  extern int cabac_encoding;
//  if (cabac_encoding)
//    fprintf(p_trace, "%d  0x%04x\n", binCount++, Erange );
//#endif

  if (symbol != 0)
    low += eep->Erange;

  /* renormalisation as for biari_encode_symbol;
     note that low has already been doubled */
  if (low >= ONE)
  {
    put_one_bit_plus_outstanding(eep,1);
    low -= ONE;
  }
  else
    if (low < HALF)
    {
      put_one_bit_plus_outstanding(eep,0);
    }
    else
    {
      eep->Ebits_to_follow++;
      low -= HALF;
    }
    eep->Elow = low;
    eep->C++;
}


/*!
 ************************************************************************
 * \brief
 *    Arithmetic encoding for last symbol before termination
 ************************************************************************
 */
void biari_encode_symbol_final(EncodingEnvironment* eep, signed short symbol)
{
  register unsigned int range = eep->Erange-2;
  register unsigned int low = eep->Elow;
//
//#if (2==TRACE)
//  extern int cabac_encoding;
//  if (cabac_encoding)
//    fprintf(p_trace, "%d  0x%04x\n", binCount++, Erange);
//#endif
//
  if (symbol) {
    low += range;
    range = 2;
  }

  while (range < QUARTER)
  {
    if (low >= HALF)
    {
      put_one_bit_plus_outstanding(eep,1);
      low -= HALF;
    }
    else
      if (low < QUARTER)
      {
        put_one_bit_plus_outstanding(eep,0);
      }
      else
      {
        eep->Ebits_to_follow++;
        low -= QUARTER;
      }
      low <<= 1;
      range <<= 1;
  }
  eep->Erange = range;
  eep->Elow = low;
  eep->C++;
}


/*!
 ************************************************************************
 * \brief
 *    Initializes a given context with some pre-defined probability state
 ************************************************************************
 */
void biari_init_context (BiContextType* ctx, const int* ini,int frame,int buffer,int QuantStepSize)
{
  int pstate=0;
  int maxQuantStepSize = 10;
  if (QuantStepSize > maxQuantStepSize)
  	QuantStepSize = maxQuantStepSize;
  
  
  //gjres avhengig av rammenr i bufferet??
  pstate = ((ini[0]* ((52*QuantStepSize)/maxQuantStepSize)) >>4) + ini[1];
  
  //hy img-qp gir grovere kvantisering, qp range: [0 52]
  //pstate = ((ini[0]* max(0, img->qp)) >> 4) + ini[1];
  
  
  pstate = min (max ( 1, pstate), 126);

  if ( pstate >= 64 )
  {
    ctx->state  = (unsigned short) (pstate - 64);
    ctx->MPS    = 1;
  }
  else
  {
    ctx->state  = (unsigned short) (63 - pstate);
    ctx->MPS    = 0;
  }

  ctx->count = 0;
}



