/* Declarations */
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>

/* Definitions */
#define PERMS        0666 /* Read and write for user, group and other */
#define ERROR          -1
#define MAXNOOFCLASSES  6 /* Max no. of entropy coder classes */
#define MAXDECLEV    9999 /* Max decision level */
#define MAXNOOFTAPS   128 /* Maximum no. of filter taps */
#define MAXNOOFCHANNELS 8 /* Maximum no. of channels */
#define MAXNOOFSUBBANDS 8 /* Maximum no. of subbands */
#define NOOFSUBBANDSLUMIN 8 /* No. of subbands (luminance component) */
#define NOOFSUBBANDSCHROM 4 /* No. of subbands (chrominance component) */
#define DECIMFACTOR     4 /* Chrominance decimation factor */
#define INTERFACTORTHUMBNAIL 2 /* Chrominance thumbnail image interpolation factor */
#define MINQIND      -128 /* Minimum quantization index */
#define MAXQIND       127 /* Maximum quantization index */
#define QINDRANGE     256 /* Maximum quantization index dynamic range */
#define OFFSET          3 /* No. of transmitted quantization repr. levels */
#define MAXK          256 /* Max size of source alphabet */
#define MAXLEN      65536 /* Max length of source string (main info) */
#define MAXLENS     65536 /* Max length of source string (side info) */
#define MAXLENS2D   65536 /* Max length of source string (side info) 2-D */
#define MAXLENS3D   16386 /* Max length of source string (side info) 3-D */
#define MAXCBITS    65536 /* Max length of code string */
#define MAXCBITSS   16384 /* Max length of code string (side info) */
#define ZERO        1e-15 /* Zero */
#define MAXPR          16 /* Max precision */
#define MAXCCW         15 /* Max width of carry-control field */
#define FILTERLENLUMIN 32 /* Length of luminance filters */
#define FILTERLENCHROM 16 /* Length of chrominance filters */
#define FILTERLENDECIM 16 /* Length of decimation filters */
#define MAXCODELENGTH 65536 /* Max length of code string */
#define max(x,y) ((x)>(y) ? x : y)
#define min(x,y) ((x)<(y) ? x : y)

#define NUM_MB_TYPE_CTX  11	/*Number of MB Type Contexts*/
#define NUM_MAP_CTX   15
#define NUM_LAST_CTX  15
#define NUM_ONE_CTX    5
#define NUM_ABS_CTX    5
#define NUM_BLOCK_TYPES 10
#define COEFFS_IN_MACROBLOCK 64
typedef unsigned char byte;

/* Structures */

//MacroBlock
typedef struct macroblock{
	int has_none_zero_coeffs;
	struct macroblock *mb_available_up, *mb_available_left;
	int mb_number;
	int bitcounter;
	short int coeffs[63];
	int position_x, position_y;
	int last_significant_coeff;	
	int coeff_cnt;	
} MacroBlock;

//EncodingEnvironment
typedef struct 
{
  unsigned int  Elow, Erange;
  unsigned int  Ebuffer;
  unsigned int  Ebits_to_go;
  unsigned int  Ebits_to_follow;
  unsigned char *Ecodestrm;
  int           *Ecodestrm_len;
  int           C;
  int           E;  
} EncodingEnvironment;

//DecodingEnvironment
typedef struct
{
  unsigned int    Dlow, Drange;
  unsigned int    Dvalue;
  unsigned int    Dbuffer;
  int             Dbits_to_go;
  byte            *Dcodestrm;
  int             *Dcodestrm_len;
} DecodingEnvironment;

//BiContextType
typedef struct  {
  unsigned short state;         // index into state-table CP  
  unsigned char  MPS;           // Least Probable Symbol 0/1 CP
  unsigned long  count;
} BiContextType;

//TextureInfoContexts
typedef struct {
  BiContextType map_contexts [NUM_BLOCK_TYPES][NUM_MAP_CTX];
  BiContextType last_contexts[NUM_BLOCK_TYPES][NUM_LAST_CTX];
  BiContextType one_contexts [NUM_BLOCK_TYPES][NUM_ONE_CTX];
  BiContextType abs_contexts [NUM_BLOCK_TYPES][NUM_ABS_CTX];
} TextureInfoContexts;

//MotionInfoContexts
typedef struct 
{  
  BiContextType mb_type_contexts[3][NUM_MB_TYPE_CTX];
} MotionInfoContexts;

//BitStream
typedef struct
{
  int             byte_pos;           //!< current position in bitstream;
  int             bits_to_go;         //!< current bitcounter
  byte            byte_buf;           //!< current buffer for last written byte
  int             stored_byte_pos;    //!< storage for position in bitstream;
  int             stored_bits_to_go;  //!< storage for bitcounter
  byte            stored_byte_buf;    //!< storage for buffer of last written byte

  byte            byte_buf_skip;      //!< current buffer for last written byte
  int             byte_pos_skip;      //!< storage for position in bitstream;
  int             bits_to_go_skip;    //!< storage for bitcounter

  byte            *streamBuffer;      //!< actual buffer for written bytes
  int             write_flag;         //!< Bitstream contains data and needs to be written

} Bitstream;

//DataPartition
typedef struct encoder_datapartition
{
  Bitstream           *bitstream;
  EncodingEnvironment ee_cabac; 
} Encoder_DataPartition;

typedef struct decoder_datapartition
{
  Bitstream           *bitstream;
  DecodingEnvironment de_cabac; 
} Decoder_DataPartition;


typedef struct {
  Encoder_DataPartition *partArr;     //!< array of partitions
  MotionInfoContexts    *mot_ctx;     //!< pointer to struct of context models for use in CABAC
  TextureInfoContexts   *tex_ctx;     //!< pointer to struct of context models for use in CABAC	
} CABAC_EncoderEnvironment;

typedef struct {
  Decoder_DataPartition *partArr;     //!< array of partitions
  MotionInfoContexts    *mot_ctx;     //!< pointer to struct of context models for use in CABAC
  TextureInfoContexts   *tex_ctx;     //!< pointer to struct of context models for use in CABAC	
} CABAC_DecoderEnvironment;

struct EncoderArgs {
	unsigned height;
	unsigned width;
	unsigned height_old;
	unsigned width_old;
	float quality;
	float threshold;
	int buffer;
	unsigned long codesize;
	int TotNoOFFrames;
	unsigned char levels;
	float LPquant;
	float LPthreshold;
};


struct DecoderArgs {
	unsigned height;
	unsigned width;
	unsigned height_old;
	unsigned width_old;
	float quality;
	float threshold;
	int buffer;
	unsigned long codesize;
	int TotNoOfFrames;
	unsigned char levels;
	float LPquant;
	float LPthreshold;
};


//biariencode.c
void put_one_bit(EncodingEnvironment*,byte);
void put_byte(EncodingEnvironment*); 
void put_one_bit_plus_outstanding(EncodingEnvironment*,byte);
void reset_pic_bin_count();
int get_pic_bin_count();
EncodingEnvironment* arienco_create_encoding_environment();
void arienco_delete_encoding_environment(EncodingEnvironment*);
void arienco_start_encoding(EncodingEnvironment*,unsigned char*,int*);
int arienco_bits_written(EncodingEnvironment*);
void arienco_done_encoding(EncodingEnvironment*);
void biari_encode_symbol(EncodingEnvironment*,signed short,BiContextType*);
void biari_encode_symbol_eq_prob(EncodingEnvironment*,signed short);
void biari_encode_symbol_final(EncodingEnvironment*,signed short);
void biari_init_context (BiContextType*,const int*,int,int,int);

//biaridecode.c
void get_byte(DecodingEnvironment*);
DecodingEnvironment* arideco_create_decoding_environment();
void arideco_delete_decoding_environment(DecodingEnvironment*);
void arideco_start_decoding(DecodingEnvironment*,unsigned char*,int,int*,int);
int arideco_bits_read(DecodingEnvironment*);
void arideco_done_decoding(DecodingEnvironment*);
unsigned int biari_decode_symbol(DecodingEnvironment*,BiContextType*);
unsigned int biari_decode_symbol_eq_prob(DecodingEnvironment*);
unsigned int biari_decode_final(DecodingEnvironment*);


//EncoderMacroblock.c
void write_macroblock (MacroBlock*,CABAC_EncoderEnvironment*,int,int,int,int);
void write_one_macroblock(MacroBlock*,TextureInfoContexts*,MotionInfoContexts*,EncodingEnvironment*);
void writeLumaCoeff8x8_CABAC(MacroBlock*,TextureInfoContexts*,MotionInfoContexts*,EncodingEnvironment*);
void write_terminating_bit (short,CABAC_EncoderEnvironment*);

//macroblock_decoder.c
void read_macroblock (MacroBlock* mb,CABAC_DecoderEnvironment*,int,int,int);
void read_one_macroblock(MacroBlock*,TextureInfoContexts*,MotionInfoContexts*,DecodingEnvironment*);
void readLumaCoeff8x8_CABAC(MacroBlock*,TextureInfoContexts*,MotionInfoContexts*,DecodingEnvironment*);
int exit_macroblocks(DecodingEnvironment*);


//cabac.c
void writeMB_skip_flagInfo_CABAC(EncodingEnvironment*,MacroBlock*,MotionInfoContexts*);
void write_significant_coefficients (MacroBlock*, EncodingEnvironment*,TextureInfoContexts*,int,short int[]);
void unary_exp_golomb_level_encode( EncodingEnvironment*,unsigned int,BiContextType*);
void exp_golomb_encode_eq_prob(EncodingEnvironment*,unsigned int,int);
void write_significance_map (MacroBlock*,EncodingEnvironment*,TextureInfoContexts*,int,short int[],int);

//cabac_decoder.c
int read_significance_map (MacroBlock*,DecodingEnvironment*,TextureInfoContexts*,int,short int[]);
void readMB_skip_flagInfo_CABAC(DecodingEnvironment*,MacroBlock*,MotionInfoContexts*);
void read_significant_coefficients (MacroBlock*,DecodingEnvironment*,TextureInfoContexts*,int,short int[]);
unsigned int unary_bin_decode(DecodingEnvironment*,BiContextType*,int);
unsigned int unary_bin_max_decode(DecodingEnvironment*,BiContextType*,int,unsigned int);
unsigned int unary_exp_golomb_level_decode(DecodingEnvironment*,BiContextType*);
unsigned int unary_exp_golomb_mv_decode(DecodingEnvironment*,BiContextType*,unsigned int);


//context_ini.c
void create_context_memory (int);
void free_context_memory ();
int SetCtxModelNumber (int,int,int);
void init_contexts (MotionInfoContexts*,TextureInfoContexts*,int,int,int,int);
double XRate (BiContextType*, const int*);
void GetCtxModelNumber (int*,MotionInfoContexts*,TextureInfoContexts*);
void store_contexts (int,int,int,int,MotionInfoContexts*,TextureInfoContexts*);
void update_field_frame_contexts (int);

//alloc_CABAC.c
CABAC_EncoderEnvironment* init_encoder_environment(unsigned char**,int*);
CABAC_DecoderEnvironment* init_decoder_environment(unsigned char**,int*);
void delete_decoder_environment(CABAC_DecoderEnvironment*);
void delete_encoder_environment(CABAC_EncoderEnvironment*);
MotionInfoContexts* create_contexts_MotionInfo(void);
TextureInfoContexts* create_contexts_TextureInfo(void);
Encoder_DataPartition* create_encoder_dataPartition(unsigned char**,int*);
Decoder_DataPartition* create_decoder_dataPartition(unsigned char**,int*);
void delete_contexts_MotionInfo(MotionInfoContexts*);
void delete_contexts_TextureInfo(TextureInfoContexts*);
void delete_encoder_DataPartition(Encoder_DataPartition*);
void delete_decoder_DataPartition(Decoder_DataPartition*);


/*Zigzag-scan functions*/
int zig_zag_scan(MacroBlock*,short int**,int,int,int);
void inv_zig_zag_scan(MacroBlock*,short int**,int,int,int);

/* ImageEncode.c */
int ImageEncode(float**,unsigned char**,struct EncoderArgs*, FILE*,int, int,CABAC_EncoderEnvironment*,int*,double*);
void write_All_Encoder_Parameters(struct EncoderArgs*,unsigned char*,int*);


/* unifquant.c */
void unifquant(float**,short int**,int,int,int,float,float,unsigned char*,int*,int,int,int);
void Quantization(float**,short int**,int,int,float,float,int,short int);
short int Unif_Quant(float,float,float,short int);

//EZW
void ezw_encode(float**,int,int,int,int,int,float,float,unsigned char*,int*); 
void ezw_decode(float**,int,int,int,int,int,float,float,unsigned char*,int*);

/* ImageDecode.c */
int ImageDecode(float**,unsigned char**,int,struct DecoderArgs*,FILE*,CABAC_DecoderEnvironment*);
void Read_All_Decoder_Parameters(struct DecoderArgs*,unsigned char*,int*);

/* invunifquant.c */
void invunifquant(float**,short int**,int,int,int,int,int);
void Inv_Quantization(short int**,float**,int,int,float,float,int);
float Inv_Unif_Quant(short int,float,float);
