#include "codecdecl.h"
#include "util.h"
#include "alloc.h"
#include <math.h>
#include <malloc.h>

void unifquant(float **SubbandImage,short int **QuantIndexArray,int RowsImage,int ColsImage,int NoOfSubbands,float QuantStepSize,float Threshold,unsigned char *Code,int *CodeCnt,int frame,int buffer,int levels)
{
	short int No_Of_Bits = 11;
	
	/* Uniform midtread quantization of all subbands (except  */
	/* the LP-LP band which is EZW coded) */
	Quantization(SubbandImage,QuantIndexArray,RowsImage,ColsImage,QuantStepSize,Threshold,NoOfSubbands,No_Of_Bits);		
}  

void Quantization(float **SubbandImage,short int **QuantIndexArray, int RowsImage,int ColsImage,float QuantStepSize,float Threshold,int NoOfSubbands,short int No_Of_Bits)
{  /* BEGIN */
   int i, j; //Index;   
	
	/* Uniform midtread quantization */	
   for (i=0;i<RowsImage;i++){
      for (j=0;j<ColsImage;j++){
			/* If not in LP-LP band */
			if ((i>=RowsImage/NoOfSubbands) || (j>=ColsImage/NoOfSubbands)){
				/* Higher bands: Assume zero mean */ 
				QuantIndexArray[i][j]=Unif_Quant(SubbandImage[i][j],QuantStepSize,Threshold,No_Of_Bits);
			}
		}	
	}
}  /* END */

short int Unif_Quant(float x, float QuantStepSize, float Threshold, short int No_Of_Quant_Bits)
{
   short int Index;    
   short int MaxIndex = (1<<(No_Of_Quant_Bits-1))-1;     

   /* Test if x is less than the minimum decision level */
   if (x < -MaxIndex*QuantStepSize-Threshold){
		Index=-MaxIndex;
		fprintf(stderr,"\nUnderflow: Quantizer dynamic range exceeded!\n");
		fprintf(stderr,"x=%6.2f < %6.2f\n",x,-(float)MaxIndex*QuantStepSize-Threshold);								
		return Index;
   }
   
   /* Test if x is larger than the maximum decision level */
   if (x >= MaxIndex*QuantStepSize+Threshold){
		Index=MaxIndex;
		fprintf(stderr,"\nOverflow: Quantizer dynamic range exceeded!\n");
		fprintf(stderr,"x=%6.2f >= %6.2f\n",x,(float)MaxIndex*QuantStepSize+Threshold);				
		return Index;
	}
   
	/* x is larger than the least decision level or less than the */
	/* largest decision level */
	if (fabs(x) < Threshold){ /* Thresholding */
		Index=0;
	} 
	else if (x <= -Threshold){
		Index = (short int)floor((x+Threshold)/QuantStepSize);
		//Index=-MINQIND+floor((x+Threshold)/QuantStepSize);		
	} 
	else if (x >= Threshold){
		Index=(short int)floor((x-Threshold)/QuantStepSize)+1;		
	}
	else {
		fprintf(stderr,"Error in quantization: Could not assign index to x = %f\n",x);
		exit(ERROR);
	}   
	return(Index);
}
